"""SSL Ciphers

Copyright (c) 1999-2003 Ng Pheng Siong. All rights reserved."""

__all__ = ["Cipher", "Cipher_Stack"]

from M2Crypto import m2, types as C
from typing import Iterable  # noqa


class Cipher(object):
    cipher: C.SSL_CIPHER

    def __init__(self, cipher: C.SSL_CIPHER, _pyfree: int = 0) -> None:
        self.cipher = cipher

    def __len__(self) -> int:
        return m2.ssl_cipher_get_bits(self.cipher)

    def __repr__(self) -> str:
        return "%s-%s" % (self.name(), len(self))

    def __str__(self) -> str:
        return "%s-%s" % (self.name(), len(self))

    def version(self) -> str:
        return m2.ssl_cipher_get_version(self.cipher)

    def name(self) -> str:
        return m2.ssl_cipher_get_name(self.cipher)


class Cipher_Stack(object):
    stack: C.STACK_OF_SSL_CIPHER

    def __init__(self, stack: C.STACK_OF_SSL_CIPHER, _pyfree: int = 0) -> None:
        """
        :param stack: binary of the C-type STACK_OF(SSL_CIPHER)
        """
        self.stack = stack

    def __len__(self) -> int:
        return m2.sk_ssl_cipher_num(self.stack)

    def __getitem__(self, idx: int) -> Cipher:
        if not 0 <= idx < m2.sk_ssl_cipher_num(self.stack):
            raise IndexError("index out of range")
        v = m2.sk_ssl_cipher_value(self.stack, idx)
        return Cipher(v)

    def __iter__(self) -> Iterable:
        for i in range(m2.sk_ssl_cipher_num(self.stack)):
            yield self[i]



"""
M2Crypto wrapper for OpenSSL ASN1 API.

Copyright (c) 1999-2004 Ng Pheng Siong. All rights reserved.

Portions created by Open Source Applications Foundation (OSAF) are
Copyright (C) 2005 OSAF. All Rights Reserved.
"""

import datetime
import time

from M2Crypto import BIO, m2, types as C
from typing import Optional, Union, Any

MBSTRING_FLAG: int = 0x1000
MBSTRING_ASC: int = MBSTRING_FLAG | 1
MBSTRING_BMP: int = MBSTRING_FLAG | 2


class ASN1_Integer:

    def __init__(self, asn1int: Union[C.ASN1_Integer, int], _pyfree: int = 0):
        if isinstance(asn1int, int):
            self.asn1int: C.ASN1_Integer = m2.asn1_integer_new()
            m2.asn1_integer_set(self.asn1int, asn1int)
        else:
            self.asn1int = asn1int
        self._pyfree = _pyfree

    def __cmp__(self, other: "ASN1_Integer") -> int:
        if not isinstance(other, ASN1_Integer):
            raise TypeError("Comparisons supported only between ANS1_Integer objects")

        return m2.asn1_integer_cmp(self.asn1int, other.asn1int)

    def __del__(self) -> None:
        if self._pyfree:
            m2.asn1_integer_free(self.asn1int)

    def __int__(self) -> int:
        ret = m2.asn1_integer_get(self.asn1int)
        if ret is None:
            raise ValueError("ASN1_Integer value is NULL")
        return ret


class ASN1_String:

    def __init__(
        self,
        asn1str: Union[C.ASN1_String, str, bytes],
        _pyfree: int = 0,
    ):
        self.asn1str: C.ASN1_String = m2.asn1_string_new()
        if isinstance(asn1str, str):
            encoded_str = asn1str.encode()
            m2.asn1_string_set(self.asn1str, encoded_str)
        elif isinstance(asn1str, bytes):
            m2.asn1_string_set(self.asn1str, asn1str)
        else:
            self.asn1str = asn1str
        self._pyfree = _pyfree

    def __bytes__(self) -> bytes:
        buf = BIO.MemoryBuffer()
        m2.asn1_string_print(buf.bio_ptr(), self.asn1str)
        return buf.read_all()

    def __str__(self) -> str:
        return self.__bytes__().decode()

    def __del__(self) -> None:
        if getattr(self, "_pyfree", 0):
            m2.asn1_string_free(self.asn1str)

    def _ptr(self):
        return self.asn1str

    def as_text(self, flags: int = 0) -> str:
        """Output an ASN1_STRING structure according to the set flags.

        :param flags: determine the format of the output by using
               predetermined constants, see ASN1_STRING_print_ex(3)
               manpage for their meaning.
        :return: output an ASN1_STRING structure.
        """
        buf = BIO.MemoryBuffer()
        m2.asn1_string_print_ex(buf.bio_ptr(), self.asn1str, flags)
        return buf.read_all().decode()


class ASN1_Object:

    m2_asn1_object_free = m2.asn1_object_free

    def __init__(self, asn1obj: C.ASN1_Object, _pyfree: int = 0) -> None:
        self.asn1obj = asn1obj
        self._pyfree = _pyfree

    def __del__(self) -> None:
        if self._pyfree:
            m2.asn1_object_free(self.asn1obj)

    def _ptr(self):
        return self.asn1obj


class _UTC(datetime.tzinfo):
    def tzname(self, dt: Optional[datetime.datetime]) -> str:
        return "UTC"

    def dst(self, dt: Optional[datetime.datetime]) -> datetime.timedelta:
        return datetime.timedelta(0)

    def utcoffset(self, dt: Optional[datetime.datetime]) -> datetime.timedelta:
        return datetime.timedelta(0)

    def __repr__(self) -> str:
        return "<Timezone: %s>" % self.tzname(None)


UTC: _UTC = _UTC()


class LocalTimezone(datetime.tzinfo):
    """Localtimezone from datetime manual."""

    def __init__(self) -> None:
        self._stdoffset = datetime.timedelta(seconds=-time.timezone)
        if time.daylight:
            self._dstoffset = datetime.timedelta(seconds=-time.altzone)
        else:
            self._dstoffset = self._stdoffset
        self._dstdiff = self._dstoffset - self._stdoffset

    def utcoffset(self, dt: Optional[datetime.datetime]) -> datetime.timedelta:
        if self._isdst(dt):
            return self._dstoffset
        else:
            return self._stdoffset

    def dst(self, dt: Optional[datetime.datetime]) -> datetime.timedelta:
        if self._isdst(dt):
            return self._dstdiff
        else:
            return datetime.timedelta(0)

    def tzname(self, dt: Optional[datetime.datetime]) -> str:
        if dt is None:
            return ""
        return time.tzname[self._isdst(dt).real]

    def _isdst(self, dt: Optional[datetime.datetime]) -> bool:
        if dt is None:
            return False
        tt = (
            dt.year,
            dt.month,
            dt.day,
            dt.hour,
            dt.minute,
            dt.second,
            dt.weekday(),
            0,
            -1,
        )
        stamp = time.mktime(tt)
        tt = time.localtime(stamp)
        return tt.tm_isdst > 0


class ASN1_TIME:
    _ssl_months = [
        "Jan",
        "Feb",
        "Mar",
        "Apr",
        "May",
        "Jun",
        "Jul",
        "Aug",
        "Sep",
        "Oct",
        "Nov",
        "Dec",
    ]
    m2_asn1_time_free = m2.asn1_time_free

    def __init__(
        self,
        asn1_time: Optional[C.ASN1_Time] = None,
        _pyfree: int = 0,
        asn1_utctime: Optional[C.ASN1_Time] = None,
    ):
        # handle old keyword parameter
        if asn1_time is None:
            asn1_time = asn1_utctime
        if asn1_time is not None:
            assert m2.asn1_time_type_check(asn1_time), "'asn1_time' type error'"
            self.asn1_time = asn1_time
            self._pyfree = _pyfree
        else:
            # that's (ASN1_TIME*)
            self.asn1_time: bytes = m2.asn1_time_new()  # type: ignore [no-redef]
            self._pyfree = 1
        self.owner: Any = None

    def __del__(self) -> None:
        if getattr(self, "_pyfree", 0):
            m2.asn1_time_free(self.asn1_time)

    def __str__(self) -> str:
        assert m2.asn1_time_type_check(self.asn1_time), "'asn1_time' type error'"
        buf = BIO.MemoryBuffer()
        m2.asn1_time_print(buf.bio_ptr(), self.asn1_time)
        return buf.read_all().decode()

    def _ptr(self):
        assert m2.asn1_time_type_check(self.asn1_time), "'asn1_time' type error'"
        return self.asn1_time

    def set_string(self, string: str) -> int:
        """
        Set time from UTC string.

        :return:  1 if the time value is successfully set and 0
                  otherwise
        """
        assert m2.asn1_time_type_check(self.asn1_time), "'asn1_time' type error'"
        return m2.asn1_time_set_string(self.asn1_time, string)

    def set_time(self, time: int) -> C.ASN1_Time:
        """
        Set time from seconds since epoch (int).

        :return: pointer to a time structure or NULL if an error
                 occurred
        """
        assert m2.asn1_time_type_check(self.asn1_time), "'asn1_time' type error'"
        return m2.asn1_time_set(self.asn1_time, time)

    def get_datetime(self) -> datetime.datetime:
        """
        Get time as datetime.datetime object

        :return: always return datetime object
        :raises: ValueError if anything wrong happens
        """
        date = str(self)

        timezone = None
        if " " not in date:
            raise ValueError("Invalid date: %s" % date)
        month, rest = date.split(" ", 1)
        if month not in self._ssl_months:
            raise ValueError("Invalid date %s: Invalid month: %s" % (date, month))
        if rest.endswith(" GMT"):
            timezone = UTC
            rest = rest[:-4]
        if "." in rest:
            dt = datetime.datetime.strptime(rest, "%d %H:%M:%S.%f %Y")
        else:
            dt = datetime.datetime.strptime(rest, "%d %H:%M:%S %Y")
        dt = dt.replace(month=self._ssl_months.index(month) + 1)
        if timezone:
            dt = dt.replace(tzinfo=UTC)
        return dt

    def set_datetime(self, date: datetime.datetime) -> C.ASN1_Time:
        """
        Set time from datetime.datetime object.

        :return: pointer to a time structure or NULL if an error
                 occurred
        """
        local = LocalTimezone()
        if date.tzinfo is None:
            date = date.replace(tzinfo=local)
        date = date.astimezone(local)
        return self.set_time(int(time.mktime(date.timetuple())))


ASN1_UTCTIME = ASN1_TIME



"""Secure Authenticator Cookies

Copyright (c) 1999-2002 Ng Pheng Siong. All rights reserved."""

import logging
import re
import time

from http.cookies import SimpleCookie

from M2Crypto import Rand, m2, util

from typing import Pattern, Optional, Tuple, Union  # noqa

_MIX_FORMAT = "exp=%f&data=%s&digest="
_MIX_RE = re.compile(r"exp=(\d+\.\d+)&data=(.+)&digest=(\S*)")

log = logging.getLogger(__name__)


def mix(expiry: float, data: str, format: str = _MIX_FORMAT) -> str:
    return format % (expiry, data)


def unmix(
    dough: Union[str, bytes], regex: Pattern = _MIX_RE
) -> Optional[Tuple[float, str]]:
    mo = regex.match(dough)
    if mo:
        return float(mo.group(1)), mo.group(2)
    else:
        return None


def unmix3(
    dough: Union[str, bytes], regex: Pattern = _MIX_RE
) -> Optional[Tuple[float, str, str]]:
    mo = regex.match(dough)
    if mo:
        return float(mo.group(1)), mo.group(2), mo.group(3)
    else:
        return None


_TOKEN: str = "_M2AUTH_"


class AuthCookie:

    def __init__(self, expiry: float, data: str, dough: str, mac: str) -> None:
        """
        Create new authentication cookie

        :param expiry: expiration time (in seconds)
        :param data: cookie payload (as a string)
        :param dough: expiry & data concatenated to URL compliant
                      string
        :param mac: SHA1-based HMAC of dough and random key
        """
        self._expiry = expiry
        self._data = data
        self._mac = mac
        self._cookie = SimpleCookie()
        self._cookie[_TOKEN] = "%s%s" % (dough, mac)
        self._name = "%s%s" % (dough, mac)  # WebKit only.

    def expiry(self) -> float:
        """Return the cookie's expiry time."""
        return self._expiry

    def data(self) -> str:
        """Return the data portion of the cookie."""
        return self._data

    def mac(self) -> str:
        """Return the cookie's MAC."""
        return self._mac

    def output(self, header: str = "Set-Cookie:") -> str:
        """Return the cookie's output in "Set-Cookie" format."""
        return self._cookie.output(header=header)

    def value(self) -> str:
        """Return the cookie's output minus the "Set-Cookie: " portion."""
        return self._cookie[_TOKEN].value

    def isExpired(self) -> bool:
        """Return 1 if the cookie has expired, 0 otherwise."""
        return isinstance(self._expiry, (float, int)) and (time.time() > self._expiry)

    # Following two methods are for WebKit only.
    # I may wish to push them to WKAuthCookie, but they are part
    # of the API now. Oh well.
    def name(self) -> str:
        return self._name

    def headerValue(self) -> str:
        return self.value()


class AuthCookieJar:

    _keylen: int = 20

    def __init__(self) -> None:
        self._key = Rand.rand_bytes(self._keylen)

    def _hmac(self, key: bytes, data: str) -> str:
        return util.bin_to_hex(m2.hmac(key, data.encode(), m2.sha256()))

    def makeCookie(self, expiry: float, data: str) -> AuthCookie:
        """
        Make a cookie

        :param expiry: expiration time (float in seconds)
        :param data: cookie content
        :return: AuthCookie object
        """
        if not isinstance(expiry, (int, float)):
            raise ValueError('Expiration time must be number, not "%s' % expiry)
        dough = mix(expiry, data)
        return AuthCookie(expiry, data, dough, self._hmac(self._key, dough))

    def isGoodCookie(self, cookie: AuthCookie) -> bool:
        if cookie.isExpired():
            return False
        c = self.makeCookie(cookie._expiry, cookie._data)
        return (
            (c._expiry == cookie._expiry)
            and (c._data == cookie._data)
            and (c._mac == cookie._mac)
            and (c.output() == cookie.output())
        )

    def isGoodCookieString(
        self, cookie_str: Union[dict, str], _debug: bool = False
    ) -> bool:
        c = SimpleCookie()
        c.load(cookie_str)
        if _TOKEN not in c:
            log.debug("_TOKEN not in c (keys = %s)", dir(c))
            return False
        undough = unmix3(c[_TOKEN].value)
        if undough is None:
            log.debug("undough is None")
            return False
        exp, data, mac = undough
        c2 = self.makeCookie(exp, data)
        if _debug and (c2._mac == mac):
            log.error("cookie_str = %s", cookie_str)
            log.error("c2.isExpired = %s", c2.isExpired())
            log.error("mac = %s", mac)
            log.error("c2._mac = %s", c2._mac)
            log.error("c2._mac == mac: %s", str(c2._mac == mac))
        return (not c2.isExpired()) and (c2._mac == mac)



"""
M2Crypto wrapper for OpenSSL ECDH/ECDSA API.

@requires: OpenSSL 0.9.8 or newer

Copyright (c) 1999-2003 Ng Pheng Siong. All rights reserved.

Portions copyright (c) 2005-2006 Vrije Universiteit Amsterdam.
All rights reserved."""

from typing import Callable, Dict, List, NoReturn, Optional, Tuple, Union  # noqa
from M2Crypto import BIO, Err, EVP, m2, types as C, util

EC_Key = bytes


class ECError(Exception):
    pass


m2.ec_init(ECError)

# Curve identifier constants
NID_secp112r1: int = m2.NID_secp112r1
NID_secp112r2: int = m2.NID_secp112r2
NID_secp128r1: int = m2.NID_secp128r1
NID_secp128r2: int = m2.NID_secp128r2
NID_secp160k1: int = m2.NID_secp160k1
NID_secp160r1: int = m2.NID_secp160r1
NID_secp160r2: int = m2.NID_secp160r2
NID_secp192k1: int = m2.NID_secp192k1
NID_secp224k1: int = m2.NID_secp224k1
NID_secp224r1: int = m2.NID_secp224r1
NID_secp256k1: int = m2.NID_secp256k1
NID_secp384r1: int = m2.NID_secp384r1
NID_secp521r1: int = m2.NID_secp521r1
NID_sect113r1: int = m2.NID_sect113r1
NID_sect113r2: int = m2.NID_sect113r2
NID_sect131r1: int = m2.NID_sect131r1
NID_sect131r2: int = m2.NID_sect131r2
NID_sect163k1: int = m2.NID_sect163k1
NID_sect163r1: int = m2.NID_sect163r1
NID_sect163r2: int = m2.NID_sect163r2
NID_sect193r1: int = m2.NID_sect193r1
NID_sect193r2: int = m2.NID_sect193r2
# default for secg.org TLS test server
NID_sect233k1: int = m2.NID_sect233k1
NID_sect233r1: int = m2.NID_sect233r1
NID_sect239k1: int = m2.NID_sect239k1
NID_sect283k1: int = m2.NID_sect283k1
NID_sect283r1: int = m2.NID_sect283r1
NID_sect409k1: int = m2.NID_sect409k1
NID_sect409r1: int = m2.NID_sect409r1
NID_sect571k1: int = m2.NID_sect571k1
NID_sect571r1: int = m2.NID_sect571r1

NID_prime192v1: int = m2.NID_X9_62_prime192v1
NID_prime192v2: int = m2.NID_X9_62_prime192v2
NID_prime192v3: int = m2.NID_X9_62_prime192v3
NID_prime239v1: int = m2.NID_X9_62_prime239v1
NID_prime239v2: int = m2.NID_X9_62_prime239v2
NID_prime239v3: int = m2.NID_X9_62_prime239v3
NID_prime256v1: int = m2.NID_X9_62_prime256v1
NID_c2pnb163v1: int = m2.NID_X9_62_c2pnb163v1
NID_c2pnb163v2: int = m2.NID_X9_62_c2pnb163v2
NID_c2pnb163v3: int = m2.NID_X9_62_c2pnb163v3
NID_c2pnb176v1: int = m2.NID_X9_62_c2pnb176v1
NID_c2tnb191v1: int = m2.NID_X9_62_c2tnb191v1
NID_c2tnb191v2: int = m2.NID_X9_62_c2tnb191v2
NID_c2tnb191v3: int = m2.NID_X9_62_c2tnb191v3
NID_c2pnb208w1: int = m2.NID_X9_62_c2pnb208w1
NID_c2tnb239v1: int = m2.NID_X9_62_c2tnb239v1
NID_c2tnb239v2: int = m2.NID_X9_62_c2tnb239v2
NID_c2tnb239v3: int = m2.NID_X9_62_c2tnb239v3
NID_c2pnb272w1: int = m2.NID_X9_62_c2pnb272w1
NID_c2pnb304w1: int = m2.NID_X9_62_c2pnb304w1
NID_c2tnb359v1: int = m2.NID_X9_62_c2tnb359v1
NID_c2pnb368w1: int = m2.NID_X9_62_c2pnb368w1
NID_c2tnb431r1: int = m2.NID_X9_62_c2tnb431r1

# To preserve compatibility with older names
NID_X9_62_prime192v1: int = NID_prime192v1
NID_X9_62_prime192v2: int = NID_prime192v2
NID_X9_62_prime192v3: int = NID_prime192v3
NID_X9_62_prime239v1: int = NID_prime239v1
NID_X9_62_prime239v2: int = NID_prime239v2
NID_X9_62_prime239v3: int = NID_prime239v3
NID_X9_62_prime256v1: int = NID_prime256v1
NID_X9_62_c2pnb163v1: int = NID_c2pnb163v1
NID_X9_62_c2pnb163v2: int = NID_c2pnb163v2
NID_X9_62_c2pnb163v3: int = NID_c2pnb163v3
NID_X9_62_c2pnb176v1: int = NID_c2pnb176v1
NID_X9_62_c2tnb191v1: int = NID_c2tnb191v1
NID_X9_62_c2tnb191v2: int = NID_c2tnb191v2
NID_X9_62_c2tnb191v3: int = NID_c2tnb191v3
NID_X9_62_c2pnb208w1: int = NID_c2pnb208w1
NID_X9_62_c2tnb239v1: int = NID_c2tnb239v1
NID_X9_62_c2tnb239v2: int = NID_c2tnb239v2
NID_X9_62_c2tnb239v3: int = NID_c2tnb239v3
NID_X9_62_c2pnb272w1: int = NID_c2pnb272w1
NID_X9_62_c2pnb304w1: int = NID_c2pnb304w1
NID_X9_62_c2tnb359v1: int = NID_c2tnb359v1
NID_X9_62_c2pnb368w1: int = NID_c2pnb368w1
NID_X9_62_c2tnb431r1: int = NID_c2tnb431r1

NID_wap_wsg_idm_ecid_wtls1: int = m2.NID_wap_wsg_idm_ecid_wtls1
NID_wap_wsg_idm_ecid_wtls3: int = m2.NID_wap_wsg_idm_ecid_wtls3
NID_wap_wsg_idm_ecid_wtls4: int = m2.NID_wap_wsg_idm_ecid_wtls4
NID_wap_wsg_idm_ecid_wtls5: int = m2.NID_wap_wsg_idm_ecid_wtls5
NID_wap_wsg_idm_ecid_wtls6: int = m2.NID_wap_wsg_idm_ecid_wtls6
NID_wap_wsg_idm_ecid_wtls7: int = m2.NID_wap_wsg_idm_ecid_wtls7
NID_wap_wsg_idm_ecid_wtls8: int = m2.NID_wap_wsg_idm_ecid_wtls8
NID_wap_wsg_idm_ecid_wtls9: int = m2.NID_wap_wsg_idm_ecid_wtls9
NID_wap_wsg_idm_ecid_wtls10: int = m2.NID_wap_wsg_idm_ecid_wtls10
NID_wap_wsg_idm_ecid_wtls11: int = m2.NID_wap_wsg_idm_ecid_wtls11
NID_wap_wsg_idm_ecid_wtls12: int = m2.NID_wap_wsg_idm_ecid_wtls12

# The following two curves, according to OpenSSL, have a
# "Questionable extension field!" and are not supported by
# the OpenSSL inverse function.  ECError: no inverse.
# As such they cannot be used for signing.  They might,
# however, be usable for encryption but that has not
# been tested.  Until thir usefulness can be established,
# they are not supported at this time.
# NID_ipsec3 = m2.NID_ipsec3
# NID_ipsec4 = m2.NID_ipsec4


class EC(object):
    """
    Object interface to a EC key pair.
    """

    def __init__(self, ec: C.EC_KEY, _pyfree: int = 0) -> None:
        assert m2.ec_key_type_check(ec), "'ec' type error"
        self.ec = ec
        self._pyfree = _pyfree

    @staticmethod
    def m2_ec_key_free(ec: C.EC_KEY) -> None:
        m2.ec_key_free(ec)

    def __del__(self) -> None:
        if getattr(self, "_pyfree", 0):
            self.m2_ec_key_free(self.ec)

    def __len__(self) -> int:
        assert m2.ec_key_type_check(self.ec), "'ec' type error"
        return m2.ec_key_keylen(self.ec)

    def gen_key(self) -> None:
        """
        Generates the key pair from its parameters. Use::

            keypair = EC.gen_params(curve)
            keypair.gen_key()

        to create an EC key pair.
        """
        assert m2.ec_key_type_check(self.ec), "'ec' type error"
        m2.ec_key_gen_key(self.ec)

    def pub(self) -> "EC_pub":
        # Don't let python free
        return EC_pub(self.ec, 0)

    def sign_dsa(self, digest: bytes) -> Tuple[bytes, bytes]:
        """
        Sign the given digest using ECDSA. Returns a tuple (r,s), the two
        ECDSA signature parameters.
        """
        assert self._check_key_type(), "'ec' type error"
        return m2.ecdsa_sign(self.ec, digest)

    def verify_dsa(self, digest: bytes, r: bytes, s: bytes) -> int:
        """
        Verify the given digest using ECDSA. r and s are the ECDSA
        signature parameters.
        """
        assert self._check_key_type(), "'ec' type error"
        return m2.ecdsa_verify(self.ec, digest, r, s)

    def sign_dsa_asn1(self, digest: bytes) -> bytes:
        assert self._check_key_type(), "'ec' type error"
        return m2.ecdsa_sign_asn1(self.ec, digest)

    def verify_dsa_asn1(self, digest: bytes, blob: bytes) -> int:
        assert self._check_key_type(), "'ec' type error"
        return m2.ecdsa_verify_asn1(self.ec, digest, blob)

    def compute_dh_key(self, pub_key: "EC") -> Optional[bytes]:
        """
        Compute the ECDH shared key of this key pair and the given public
        key object. They must both use the same curve. Returns the
        shared key in binary as a buffer object. No Key Derivation Function is
        applied.
        """
        assert self.check_key(), "key is not initialised"
        return m2.ecdh_compute_key(self.ec, pub_key.ec)

    def save_key_bio(
        self,
        bio: BIO.BIO,
        cipher: Optional[str] = "aes_128_cbc",
        callback: Callable = util.passphrase_callback,
    ) -> int:
        """
        Save the key pair to an M2Crypto.BIO.BIO object in PEM format.

        :param bio: M2Crypto.BIO.BIO object to save key to.

        :param cipher: Symmetric cipher to protect the key. The default
                       cipher is 'aes_128_cbc'. If cipher is None, then
                       the key is saved in the clear.

        :param callback: A Python callable object that is invoked
                         to acquire a passphrase with which to protect
                         the key. The default is
                         util.passphrase_callback.
        """
        if cipher is None:
            return m2.ec_key_write_bio_no_cipher(self.ec, bio._ptr(), callback)
        else:
            ciph = getattr(m2, cipher, None)
            if ciph is None:
                raise ValueError("not such cipher %s" % cipher)
            return m2.ec_key_write_bio(self.ec, bio._ptr(), ciph(), callback)

    def save_key(
        self,
        file: Union[str, bytes],
        cipher: Optional[str] = "aes_128_cbc",
        callback: Callable = util.passphrase_callback,
    ) -> int:
        """
        Save the key pair to a file in PEM format.

        :param file: Name of filename to save key to.

        :param cipher: Symmetric cipher to protect the key. The default
                       cipher is 'aes_128_cbc'. If cipher is None, then
                       the key is saved in the clear.

        :param callback: A Python callable object that is invoked
                         to acquire a passphrase with which to protect
                         the key.  The default is
                         util.passphrase_callback.
        """
        with BIO.openfile(file, "wb") as bio:
            return self.save_key_bio(bio, cipher, callback)

    def save_pub_key_bio(self, bio: BIO.BIO) -> int:
        """
        Save the public key to an M2Crypto.BIO.BIO object in PEM format.

        :param bio: M2Crypto.BIO.BIO object to save key to.
        """
        return m2.ec_key_write_pubkey(self.ec, bio._ptr())

    def save_pub_key(self, file: Union[str, bytes]) -> int:
        """
        Save the public key to a filename in PEM format.

        :param file: Name of filename to save key to.
        """
        with BIO.openfile(file, "wb") as bio:
            return m2.ec_key_write_pubkey(self.ec, bio._ptr())

    def as_pem(
        self,
        cipher: str = "aes_128_cbc",
        callback: Callable = util.passphrase_callback,
    ) -> bytes:
        """
        Returns the key(pair) as a string in PEM format.
        If no password is passed and the cipher is set
        it exits with error
        """
        with BIO.MemoryBuffer() as bio:
            self.save_key_bio(bio, cipher, callback)
            return bio.read()

    def _check_key_type(self) -> int:
        return m2.ec_key_type_check(self.ec)

    def check_key(self) -> int:
        assert m2.ec_key_type_check(self.ec), "'ec' type error"
        return m2.ec_key_check_key(self.ec)


class EC_pub(EC):
    """
    Object interface to an EC public key.
    ((don't like this implementation inheritance))
    """

    def __init__(self, ec: C.EC_KEY, _pyfree: int = 0) -> None:
        EC.__init__(self, ec, _pyfree)
        self.der: Optional[bytes] = None

    def get_der(self) -> bytes:
        """
        Returns the public key in DER format as a buffer object.
        """
        assert self.check_key(), "key is not initialised"
        if self.der is None:
            self.der = m2.ec_key_get_public_der(self.ec)
        return self.der

    def get_key(self) -> bytes:
        """
        Returns the public key as a byte string.
        """
        assert self.check_key(), "key is not initialised"
        return m2.ec_key_get_public_key(self.ec)

    def as_pem(self):
        """
        Returns the key(pair) as a string in PEM format.
        If no password is passed and the cipher is set
        it exits with error
        """
        with BIO.MemoryBuffer() as bio:
            self.save_key_bio(bio)
            return bio.read()

    save_key: Callable = EC.save_pub_key  # type: ignore[assignment]
    save_key_bio: Callable = EC.save_pub_key_bio  # type: ignore[assignment]


def gen_params(curve: int) -> EC:
    """
    Factory function that generates EC parameters and
    instantiates a EC object from the output.

    :param curve: This is the OpenSSL nid of the curve to use.
    """
    assert curve in [
        x["NID"] for x in m2.ec_get_builtin_curves()
    ], "Elliptic curve %s is not available on this system." % m2.obj_nid2sn(curve)
    return EC(m2.ec_key_new_by_curve_name(curve), 1)


def load_key(
    file: Union[str, bytes],
    callback: Callable = util.passphrase_callback,
) -> EC:
    """
    Factory function that instantiates a EC object.

    :param file: Names the filename that contains the PEM representation
                 of the EC key pair.

    :param callback: Python callback object that will be invoked
                     if the EC key pair is passphrase-protected.
    """
    with BIO.openfile(file) as bio:
        return load_key_bio(bio, callback)


def load_key_string(string: str, callback: Callable = util.passphrase_callback) -> EC:
    """
    Load an EC key pair from a string.

    :param string: String containing EC key pair in PEM format.

    :param callback: A Python callable object that is invoked
                     to acquire a passphrase with which to unlock the
                     key. The default is util.passphrase_callback.

    :return: M2Crypto.EC.EC object.
    """
    with BIO.MemoryBuffer(string.encode()) as bio:
        return load_key_bio(bio, callback)


def load_key_bio(bio: BIO.BIO, callback: Callable = util.passphrase_callback) -> EC:
    """
    Factory function that instantiates a EC object.

    :param bio: M2Crypto.BIO object that contains the PEM
                representation of the EC key pair.

    :param callback: Python callback object that will be invoked
                     if the EC key pair is passphrase-protected.
    """
    key = m2.ec_key_read_bio(bio._ptr(), callback)
    if key is None:
        raise IOError(
            "Cannot read EC key pair from PEM file {}.".format(
                getattr(bio, "fname", "in-memory buffer")
            )
        )
    return EC(key, 1)


def load_pub_key(file: Union[str, bytes]) -> EC_pub:
    """
    Load an EC public key from filename.

    :param file: Name of filename containing EC public key in PEM
                 format.

    :return: M2Crypto.EC.EC_pub object.
    """
    with BIO.openfile(file) as bio:
        return load_pub_key_bio(bio)


def load_key_string_pubkey(
    string: str, callback: Callable = util.passphrase_callback
) -> "EVP.PKey":
    """
    Load an M2Crypto.EVP.PKey from a public key as a string.

    :param string: String containing the key in PEM format.

    :param callback: A Python callable object that is invoked
                     to acquire a passphrase with which to protect the
                     key.

    :return: M2Crypto.EVP.PKey object.
    """
    from M2Crypto.EVP import load_key_bio_pubkey

    with BIO.MemoryBuffer(string.encode()) as bio:
        return load_key_bio_pubkey(bio, callback)


def load_pub_key_bio(bio: BIO.BIO) -> EC_pub:
    """
    Load an EC public key from an M2Crypto.BIO.BIO object.

    :param bio: M2Crypto.BIO.BIO object containing EC public key in PEM
                format.

    :return: M2Crypto.EC.EC_pub object.
    """
    ec = m2.ec_key_read_pubkey(bio._ptr())
    if ec is None:
        ec_error()
    return EC_pub(ec, 1)


def ec_error() -> NoReturn:
    raise ECError(Err.get_error_message())


def pub_key_from_der(der: bytes) -> EC_pub:
    """
    Create EC_pub from DER.
    """
    return EC_pub(m2.ec_key_from_pubkey_der(der), 1)


def pub_key_from_params(curve: int, bytes: bytes) -> EC_pub:
    """
    Create EC_pub from curve name and octet string.
    """
    return EC_pub(m2.ec_key_from_pubkey_params(curve, bytes), 1)


def get_builtin_curves() -> List[Dict[str, Union[int, str]]]:
    return m2.ec_get_builtin_curves()



"""M2Crypto wrapper for OpenSSL S/MIME API.

Copyright (c) 1999-2003 Ng Pheng Siong. All rights reserved."""

from M2Crypto import BIO, EVP, Err, X509, m2, util, types as C
from typing import Callable, Optional, Tuple, Union  # noqa

PKCS7_TEXT: int = m2.PKCS7_TEXT
PKCS7_NOCERTS: int = m2.PKCS7_NOCERTS
PKCS7_NOSIGS: int = m2.PKCS7_NOSIGS
PKCS7_NOCHAIN: int = m2.PKCS7_NOCHAIN
PKCS7_NOINTERN: int = m2.PKCS7_NOINTERN
PKCS7_NOVERIFY: int = m2.PKCS7_NOVERIFY
PKCS7_DETACHED: int = m2.PKCS7_DETACHED
PKCS7_BINARY: int = m2.PKCS7_BINARY
PKCS7_NOATTR: int = m2.PKCS7_NOATTR
PKCS7_SIGNED: int = m2.PKCS7_SIGNED
PKCS7_ENVELOPED: int = m2.PKCS7_ENVELOPED
PKCS7_SIGNED_ENVELOPED: int = m2.PKCS7_SIGNED_ENVELOPED  # Deprecated
PKCS7_DATA: int = m2.PKCS7_DATA


class PKCS7_Error(Exception):
    pass


m2.pkcs7_init(PKCS7_Error)


class PKCS7(object):

    def __init__(self, pkcs7: Optional[C.PKCS7] = None, _pyfree: int = 0) -> None:
        """PKCS7 object.

        :param pkcs7: binary representation of
               the OpenSSL type PKCS7
        """
        if pkcs7 is not None:
            self.pkcs7 = pkcs7
            self._pyfree = _pyfree
        else:
            self.pkcs7 = m2.pkcs7_new()
            self._pyfree = 1

    @staticmethod
    def _m2_pkcs7_free(p7: C.PKCS7) -> None:
        m2.pkcs7_free(p7)

    def __del__(self) -> None:
        if getattr(self, "_pyfree", 0):
            self._m2_pkcs7_free(self.pkcs7)

    def _ptr(self):
        return self.pkcs7

    def type(self, text_name: int = 0) -> Union[int, str]:
        if text_name:
            return m2.pkcs7_type_sn(self.pkcs7)
        else:
            return m2.pkcs7_type_nid(self.pkcs7)

    def write(self, bio: BIO.BIO) -> int:
        return m2.pkcs7_write_bio(self.pkcs7, bio._ptr())

    def write_der(self, bio: BIO.BIO) -> int:
        return m2.pkcs7_write_bio_der(self.pkcs7, bio._ptr())

    def get0_signers(self, certs: X509.X509_Stack, flags: int = 0) -> X509.X509_Stack:
        return X509.X509_Stack(m2.pkcs7_get0_signers(self.pkcs7, certs.stack, flags), 1)


def load_pkcs7(p7file: Union[str, bytes]) -> PKCS7:
    with BIO.openfile(p7file, "r") as bio:
        p7_ptr = m2.pkcs7_read_bio(bio.bio)

    return PKCS7(p7_ptr, 1)


def load_pkcs7_der(p7file: Union[str, bytes]) -> PKCS7:
    with BIO.openfile(p7file, "rb") as bio:
        p7_ptr = m2.pkcs7_read_bio_der(bio.bio)

    return PKCS7(p7_ptr, 1)


def load_pkcs7_bio(p7_bio: BIO.BIO) -> PKCS7:
    p7_ptr = m2.pkcs7_read_bio(p7_bio._ptr())
    return PKCS7(p7_ptr, 1)


def load_pkcs7_bio_der(p7_bio: BIO.BIO) -> PKCS7:
    p7_ptr = m2.pkcs7_read_bio_der(p7_bio._ptr())
    return PKCS7(p7_ptr, 1)


def smime_load_pkcs7(p7file: Union[str, bytes]) -> Tuple[PKCS7, Optional[BIO.BIO]]:
    filename = p7file.decode() if isinstance(p7file, bytes) else p7file
    bio = m2.bio_new_file(filename, "r")

    try:
        p7_ptr, bio_ptr = m2.smime_read_pkcs7(bio)
    finally:
        m2.bio_free(bio)

    if bio_ptr is None:
        return PKCS7(p7_ptr, 1), None
    else:
        return PKCS7(p7_ptr, 1), BIO.BIO(bio_ptr, 1)


def smime_load_pkcs7_bio(p7_bio: BIO.BIO) -> Tuple[PKCS7, Optional[BIO.BIO]]:
    p7_ptr, bio_ptr = m2.smime_read_pkcs7(p7_bio._ptr())
    if p7_ptr is None:
        raise SMIME_Error(Err.get_error())
    if bio_ptr is None:
        return PKCS7(p7_ptr, 1), None
    else:
        return PKCS7(p7_ptr, 1), BIO.BIO(bio_ptr, 1)


class Cipher(object):
    """Object interface to EVP_CIPHER without all the frills of
    M2Crypto.EVP.Cipher.
    """

    def __init__(self, algo: str) -> None:
        cipher = getattr(m2, algo, None)
        if cipher is None:
            raise ValueError("unknown cipher", algo)
        self.cipher = cipher()

    def _ptr(self):
        return self.cipher


class SMIME_Error(Exception):
    pass


m2.smime_init(SMIME_Error)


# FIXME class has no __init__ method
class SMIME(object):
    def load_key(
        self,
        keyfile: Union[str, bytes],
        certfile: Union[str, bytes, None] = None,
        callback: Callable = util.passphrase_callback,
    ) -> None:
        if certfile is None:
            certfile = keyfile
        self.pkey = EVP.load_key(keyfile, callback)
        self.x509 = X509.load_cert(certfile)

    def load_key_bio(
        self,
        keybio: BIO.BIO,
        certbio: Optional[BIO.BIO] = None,
        callback: Callable = util.passphrase_callback,
    ) -> None:
        if certbio is None:
            certbio = keybio
        self.pkey = EVP.load_key_bio(keybio, callback)
        self.x509 = X509.load_cert_bio(certbio)

    def set_x509_stack(self, stack: X509.X509_Stack) -> None:
        assert isinstance(stack, X509.X509_Stack)
        self.x509_stack = stack

    def set_x509_store(self, store: X509.X509_Store) -> None:
        assert isinstance(store, X509.X509_Store)
        self.x509_store = store

    def set_cipher(self, cipher: Cipher) -> None:
        assert isinstance(cipher, Cipher)
        self.cipher = cipher

    def unset_key(self) -> None:
        del self.pkey
        del self.x509

    def unset_x509_stack(self) -> None:
        del self.x509_stack

    def unset_x509_store(self) -> None:
        del self.x509_store

    def unset_cipher(self) -> None:
        del self.cipher

    def encrypt(self, data_bio: BIO.BIO, flags: int = 0) -> PKCS7:
        if not hasattr(self, "cipher"):
            raise SMIME_Error("no cipher: use set_cipher()")
        if not hasattr(self, "x509_stack"):
            raise SMIME_Error("no recipient certs: use set_x509_stack()")

        pkcs7 = m2.pkcs7_encrypt(
            self.x509_stack._ptr(),
            data_bio._ptr(),
            self.cipher._ptr(),
            flags,
        )

        return PKCS7(pkcs7, 1)

    def decrypt(self, pkcs7: PKCS7, flags: int = 0) -> Optional[bytes]:
        if not hasattr(self, "pkey"):
            raise SMIME_Error("no private key: use load_key()")
        if not hasattr(self, "x509"):
            raise SMIME_Error("no certificate: load_key() used incorrectly?")
        blob = m2.pkcs7_decrypt(pkcs7._ptr(), self.pkey._ptr(), self.x509._ptr(), flags)
        return blob

    def sign(
        self,
        data_bio: BIO.BIO,
        flags: int = 0,
        algo: str = "sha256",
    ) -> PKCS7:
        if not hasattr(self, "pkey"):
            raise SMIME_Error("no private key: use load_key()")

        hash = getattr(m2, algo, None)

        if hash is None:
            raise SMIME_Error("no such hash algorithm %s" % algo)

        if hasattr(self, "x509_stack"):
            pkcs7 = m2.pkcs7_sign1(
                self.x509._ptr(),
                self.pkey._ptr(),
                self.x509_stack._ptr(),
                data_bio._ptr(),
                hash(),
                flags,
            )
            return PKCS7(pkcs7, 1)
        else:
            pkcs7 = m2.pkcs7_sign0(
                self.x509._ptr(),
                self.pkey._ptr(),
                data_bio._ptr(),
                hash(),
                flags,
            )
            return PKCS7(pkcs7, 1)

    def verify(
        self,
        pkcs7: PKCS7,
        data_bio: Optional[BIO.BIO] = None,
        flags: int = 0,
    ) -> Optional[bytes]:
        if not hasattr(self, "x509_stack"):
            raise SMIME_Error("no signer certs: use set_x509_stack()")
        if not hasattr(self, "x509_store"):
            raise SMIME_Error("no x509 cert store: use set_x509_store()")
        assert isinstance(pkcs7, PKCS7), "pkcs7 not an instance of PKCS7"
        p7 = pkcs7._ptr()
        if data_bio is None:
            blob = m2.pkcs7_verify0(
                p7,
                self.x509_stack._ptr(),
                self.x509_store._ptr(),
                flags,
            )
        else:
            blob = m2.pkcs7_verify1(
                p7,
                self.x509_stack._ptr(),
                self.x509_store._ptr(),
                data_bio._ptr(),
                flags,
            )
        return blob

    def write(
        self,
        out_bio: BIO.BIO,
        pkcs7: PKCS7,
        data_bio: Optional[BIO.BIO] = None,
        flags: int = 0,
    ) -> int:
        if data_bio is None:
            return m2.smime_write_pkcs7(out_bio._ptr(), pkcs7._ptr(), flags)
        else:
            return m2.smime_write_pkcs7_multi(
                out_bio._ptr(), pkcs7._ptr(), data_bio._ptr(), flags
            )


def text_crlf(text: bytes) -> bytes:
    bio_in = BIO.MemoryBuffer(text)
    bio_out = BIO.MemoryBuffer()
    if m2.smime_crlf_copy(bio_in._ptr(), bio_out._ptr()):
        return bio_out.read()
    else:
        raise SMIME_Error(Err.get_error())


def text_crlf_bio(bio_in: BIO.BIO) -> BIO.BIO:
    bio_out = BIO.MemoryBuffer()
    if m2.smime_crlf_copy(bio_in._ptr(), bio_out._ptr()):
        return bio_out
    else:
        raise SMIME_Error(Err.get_error())



"""Support for SSL socket timeouts.

Copyright (c) 1999-2003 Ng Pheng Siong. All rights reserved.

Copyright 2008 Heikki Toivonen. All rights reserved.
"""

__all__ = [
    "DEFAULT_TIMEOUT",
    "timeout",
    "struct_to_timeout",
    "struct_size",
]

import sys
import struct

from M2Crypto import m2

DEFAULT_TIMEOUT: int = 600


class timeout(object):
    sec: int
    microsec: int

    def __init__(self, sec: int = DEFAULT_TIMEOUT, microsec: int = 0) -> None:
        self.sec = sec
        self.microsec = microsec

    def pack(self) -> bytes:
        if sys.platform == "win32":
            millisec = int(self.sec * 1000 + round(float(self.microsec) / 1000))
            binstr = struct.pack("l", millisec)
        else:
            if m2.time_t_bits() == 32:
                binstr = struct.pack("ii", self.sec, self.microsec)
            else:
                binstr = struct.pack("ll", self.sec, self.microsec)
        return binstr


def struct_to_timeout(binstr: bytes) -> timeout:
    if sys.platform == "win32":
        millisec = struct.unpack("l", binstr)[0]
        # On py3, int/int performs exact division and returns float. We want
        # the whole number portion of the exact division result:
        sec = int(millisec / 1000)
        microsec = (millisec % 1000) * 1000
    else:
        (sec, microsec) = struct.unpack("ll", binstr)
    return timeout(sec, microsec)


def struct_size() -> int:
    if sys.platform == "win32":
        return struct.calcsize("l")
    else:
        return struct.calcsize("ll")



"""SSLServer

Copyright (c) 1999-2002 Ng Pheng Siong. All rights reserved."""

# M2Crypto
from M2Crypto.SSL.Connection import Connection
from M2Crypto.SSL.Context import Context  # noqa
from M2Crypto import util  # noqa
from socketserver import (
    BaseRequestHandler,
    BaseServer,
    TCPServer,
    ThreadingMixIn,
)
import os

if os.name != "nt":
    from socketserver import ForkingMixIn
import socket as std_socket
from typing import Union, Optional

__all__ = ["SSLServer", "ForkingSSLServer", "ThreadingSSLServer"]


class SSLServer(TCPServer):
    socket: Connection  # type: ignore[assignment]

    def __init__(
        self,
        server_address: util.AddrType,
        RequestHandlerClass: "BaseRequestHandler",
        ssl_context: "Context",
        bind_and_activate: bool = True,
    ) -> None:
        """
        Superclass says: Constructor. May be extended, do not override.
        This class says: Ho-hum.
        """
        BaseServer.__init__(self, server_address, RequestHandlerClass)  # type: ignore[arg-type]
        self.ssl_ctx = ssl_context
        self.socket = Connection(self.ssl_ctx)
        if bind_and_activate:
            self.server_bind()
            self.server_activate()

    def handle_request(self) -> None:
        from M2Crypto.SSL import SSLError

        request = None
        client_address = None
        try:
            request, client_address = self.get_request()
            if self.verify_request(request, client_address):
                self.process_request(request, client_address)
        except SSLError:
            self.handle_error(request, client_address)

    def handle_error(  # type: ignore[override]
        self,
        request: Optional[Union[std_socket.socket, Connection]],
        client_address: Optional[util.AddrType],
    ) -> None:
        print("-" * 40)
        import traceback

        traceback.print_exc()
        print("-" * 40)


class ThreadingSSLServer(ThreadingMixIn, SSLServer):
    pass


if os.name != "nt":

    class ForkingSSLServer(ForkingMixIn, SSLServer):
        pass



"""M2Crypto wrapper for OpenSSL RSA API.

Copyright (c) 1999-2004 Ng Pheng Siong. All rights reserved."""

import sys

from M2Crypto import BIO, Err, m2, util, types as C
from typing import Callable, IO, Optional, Tuple, Union, Any  # noqa


class RSAError(Exception):
    pass


m2.rsa_init(RSAError)

no_padding = m2.no_padding
pkcs1_padding = m2.pkcs1_padding
if hasattr(m2, "sslv23_padding"):
    sslv23_padding = m2.sslv23_padding
pkcs1_oaep_padding = m2.pkcs1_oaep_padding


class RSA(object):
    """
    RSA Key Pair.
    """

    def __init__(self, rsa: C.RSA, _pyfree: int = 0) -> None:
        """
        :param rsa: OpenSSL RSA object
        """
        self.rsa = rsa
        self._pyfree = _pyfree
        self._check_cache: Optional[int] = None

    def __del__(self) -> None:
        if getattr(self, "_pyfree", 0):
            # FIX: Call module function directly to avoid implicit 'self'
            m2.rsa_free(self.rsa)

    def __len__(self) -> int:
        return int(m2.rsa_size(self.rsa) << 3)

    def __getattr__(self, name: str) -> bytes:
        if name == "e":
            return m2.rsa_get_e(self.rsa)
        elif name == "n":
            return m2.rsa_get_n(self.rsa)
        else:
            raise AttributeError(name)

    def pub(self) -> Tuple[bytes, bytes]:
        if self.check_key() != 1:
            raise RSAError("key is not initialized")
        return m2.rsa_get_e(self.rsa), m2.rsa_get_n(self.rsa)

    def public_encrypt(self, data: bytes, padding: int) -> bytes:
        if self.check_key() != 1:
            raise RSAError("key is not initialized")
        return m2.rsa_public_encrypt(self.rsa, data, padding)

    def public_decrypt(self, data: bytes, padding: int) -> Optional[bytes]:
        # FIX: Return type is Optional[bytes] because it can fail.
        if self.check_key() != 1:
            raise RSAError("key is not initialized")
        return m2.rsa_public_decrypt(self.rsa, data, padding)

    def private_encrypt(self, data: bytes, padding: int) -> bytes:
        if self.check_key() != 1:
            raise RSAError("key is not initialized")
        return m2.rsa_private_encrypt(self.rsa, data, padding)

    def private_decrypt(self, data: bytes, padding: int) -> Optional[bytes]:
        # FIX: Return type is Optional[bytes] because it can fail.
        if self.check_key() != 1:
            raise RSAError("key is not initialized")
        return m2.rsa_private_decrypt(self.rsa, data, padding)

    def save_key_bio(
        self,
        bio: BIO.BIO,
        cipher: Optional[str] = "aes_128_cbc",
        callback: Callable[..., Any] = util.passphrase_callback,
    ) -> int:
        """
        Save the key pair to an M2Crypto.BIO.BIO object in PEM format.

        :param bio: M2Crypto.BIO.BIO object to save key to.

        :param cipher: Symmetric cipher to protect the key. The default
                       cipher is 'aes_128_cbc'. If cipher is None, then
                       the key is saved in the clear.

        :param callback: A Python callable object that is invoked
                         to acquire a passphrase with which to protect
                         the key.  The default is
                         util.passphrase_callback.
        """
        if bio.bio is None:
            raise RSAError("Uninitialized bio.bio")
        if cipher is None:
            return m2.rsa_write_key_no_cipher(self.rsa, bio.bio, callback)
        else:
            ciph = getattr(m2, cipher, None)
            if ciph is None:
                raise RSAError("not such cipher %s" % cipher)
            else:
                ciph_obj = ciph()
            return m2.rsa_write_key(self.rsa, bio.bio, ciph_obj, callback)

    def save_key(
        self,
        file: Union[str, bytes],
        cipher: Optional[str] = "aes_128_cbc",
        callback: Callable[..., Any] = util.passphrase_callback,
    ) -> int:
        """
        Save the key pair to a file in PEM format.

        :param file: Name of file to save key to.

        :param cipher: Symmetric cipher to protect the key. The default
                       cipher is 'aes_128_cbc'. If cipher is None, then
                       the key is saved in the clear.

        :param callback: A Python callable object that is invoked
                         to acquire a passphrase with which to protect
                         the key.  The default is
                         util.passphrase_callback.
        """
        with BIO.openfile(file, "wb") as bio:
            return self.save_key_bio(bio, cipher, callback)

    save_pem = save_key

    def as_pem(
        self,
        cipher: Optional[str] = "aes_128_cbc",
        callback: Callable[..., Any] = util.passphrase_callback,
    ) -> bytes:
        """
        Returns the key(pair) as a string in PEM format.
        """
        bio = BIO.MemoryBuffer()
        self.save_key_bio(bio, cipher, callback)
        return bio.read() or b""

    def save_key_der_bio(self, bio: BIO.BIO) -> int:
        """
        Save the key pair to an M2Crypto.BIO.BIO object in DER format.

        :param bio: M2Crypto.BIO.BIO object to save key to.
        """
        if bio.bio is None:
            raise RSAError("Uninitialized bio.bio")
        return m2.rsa_write_key_der(self.rsa, bio.bio)

    def save_key_der(self, file: Union[str, bytes]) -> int:
        """
        Save the key pair to a file in DER format.

        :param file: Filename to save key to
        """
        with BIO.openfile(file, "wb") as bio:
            return self.save_key_der_bio(bio)

    def save_pub_key_bio(self, bio: BIO.BIO) -> int:
        """
        Save the public key to an M2Crypto.BIO.BIO object in PEM format.

        :param bio: M2Crypto.BIO.BIO object to save key to.
        """
        if bio.bio is None:
            raise RSAError("Uninitialized bio.bio")
        return m2.rsa_write_pub_key(self.rsa, bio.bio)

    def save_pub_key(self, file: Union[str, bytes]) -> int:
        """
        Save the public key to a file in PEM format.

        :param file: Name of file to save key to.
        """
        with BIO.openfile(file, "wb") as bio:
            return self.save_pub_key_bio(bio)

    def check_key(self) -> int:
        """
        Validate RSA keys.

        It checks that p and q are in fact prime, and that n = p*q.

        :return: returns 1 if rsa is a valid RSA key, and 0 otherwise.
                 -1 is returned if an error occurs while checking the key.
                 If the key is invalid or an error occurred, the reason
                 code can be obtained using ERR_get_error(3).
        """
        if self._check_cache is not None:
            return self._check_cache
        self._check_cache = m2.rsa_check_key(self.rsa)
        return self._check_cache

    def sign_rsassa_pss(
        self, digest: bytes, algo: str = "sha256", salt_length: int = 20
    ) -> bytes:
        """
        Signs a digest with the private key using RSASSA-PSS

        :param digest: A digest created by using the digest method

        :param salt_length: The length of the salt to use

        :param algo: The hash algorithm to use
                     Legal values like 'sha1','sha224', 'sha256',
                     'ripemd160', and 'md5'.

        :return: a string which is the signature
        """
        hash_func = getattr(m2, algo, None)

        if hash_func is None:
            raise RSAError("not such hash algorithm %s" % algo)

        signature = m2.rsa_padding_add_pkcs1_pss(
            self.rsa, digest, hash_func(), salt_length
        )

        return self.private_encrypt(signature, no_padding)

    def verify_rsassa_pss(
        self,
        data: bytes,
        signature: bytes,
        algo: str = "sha256",
        salt_length: int = 20,
    ) -> int:
        """
        Verifies the signature RSASSA-PSS

        :param data: Data that has been signed

        :param signature: The signature signed with RSASSA-PSS

        :param salt_length: The length of the salt that was used

        :param algo: The hash algorithm to use
                     Legal values are for example 'sha1','sha224',
                     'sha256', 'ripemd160', and 'md5'.

        :return: 1 or 0, depending on whether the signature was
                 verified or not.
        """
        hash_func = getattr(m2, algo, None)

        if hash_func is None:
            raise RSAError("not such hash algorithm %s" % algo)

        plain_signature = self.public_decrypt(signature, no_padding)
        if plain_signature is None:
            return 0  # Decrypt failed, so verification fails

        return m2.rsa_verify_pkcs1_pss(
            self.rsa, data, plain_signature, hash_func(), salt_length
        )

    def sign(self, digest: bytes, algo: str = "sha256") -> bytes:
        """
        Signs a digest with the private key

        :param digest: A digest created by using the digest method

        :param algo: The method that created the digest.
                     Legal values like 'sha1','sha224', 'sha256',
                     'ripemd160', and 'md5'.

        :return: a string which is the signature
        """
        digest_type = getattr(m2, "NID_" + algo, None)
        if digest_type is None:
            raise ValueError("unknown algorithm", algo)

        return m2.rsa_sign(self.rsa, digest, digest_type)

    def verify(self, data: bytes, signature: bytes, algo: str = "sha256") -> int:
        """
        Verifies the signature with the public key

        :param data: Data that has been signed

        :param signature: The signature signed with the private key

        :param algo: The method use to create digest from the data
                     before it was signed.  Legal values like
                     'sha1','sha224', 'sha256', 'ripemd160', and 'md5'.

        :return: 1 or 0, depending on whether the signature was
                 verified or not.
        """
        digest_type = getattr(m2, "NID_" + algo, None)
        if digest_type is None:
            raise ValueError("unknown algorithm", algo)

        return m2.rsa_verify(self.rsa, data, signature, digest_type)

    def set_ex_data(self, index: int, data: Any) -> int:
        if self.check_key() != 1:
            raise RSAError("key is not initialized")
        return m2.rsa_set_ex_data(self.rsa, index, data)

    def get_ex_data(self, index: int) -> Any:
        if self.check_key() != 1:
            raise RSAError("key is not initialized")
        return m2.rsa_get_ex_data(self.rsa, index)


class RSA_pub(RSA):
    """
    Object interface to an RSA public key.
    """

    def __setattr__(self, name: str, value: Any) -> None:
        if name in ["e", "n"]:
            raise RSAError("use factory function new_pub_key() to set (e, n)")
        else:
            self.__dict__[name] = value

    def private_encrypt(self, *argv: Any) -> None:  # type: ignore[override]
        raise RSAError("RSA_pub object has no private key")

    def private_decrypt(self, *argv: Any) -> None:  # type: ignore[override]
        raise RSAError("RSA_pub object has no private key")

    def save_key(self, file: Union[str, bytes], *args: Any, **kw: Any) -> int:
        """
        Save public key to file.
        """
        return self.save_pub_key(file)

    def save_key_bio(self, bio: BIO.BIO, *args: Any, **kw: Any) -> int:
        """
        Save public key to BIO.
        """
        return self.save_pub_key_bio(bio)

    def check_key(self) -> int:
        return m2.rsa_check_pub_key(self.rsa)


def rsa_error() -> None:
    raise RSAError(Err.get_error_message())


def keygen_callback(p: int, n: int) -> int:
    """
    Default callback for gen_key().
    """
    ch = [".", "+", "*", "\n"]
    sys.stdout.write(ch[p])
    sys.stdout.flush()
    return 1


def gen_key(
    bits: int,
    e: int,
    callback: Callable[[int, int], int] = keygen_callback,
) -> RSA:
    """
    Generate an RSA key pair.

    :param bits: Key length, in bits.

    :param e: The RSA public exponent.

    :param callback: A Python callable object that is invoked
                     during key generation; its usual purpose is to
                     provide visual feedback. The default callback is
                     keygen_callback.

    :return: M2Crypto.RSA.RSA object.
    """
    rsa_obj = m2.rsa_generate_key(bits, e, callback)
    if rsa_obj is None:
        rsa_error()
    return RSA(rsa_obj, 1)


def load_key(
    file: Union[str, bytes],
    callback: Callable[..., Any] = util.passphrase_callback,
) -> RSA:
    """
    Load an RSA key pair from file.

    :param file: Name of file containing RSA public key in PEM format.

    :param callback: A Python callable object that is invoked
                     to acquire a passphrase with which to unlock the
                     key.  The default is util.passphrase_callback.

    :return: M2Crypto.RSA.RSA object.
    """
    with BIO.openfile(file) as bio:
        return load_key_bio(bio, callback)


def load_key_bio(
    bio: BIO.BIO, callback: Callable[..., Any] = util.passphrase_callback
) -> RSA:
    """
    Load an RSA key pair from an M2Crypto.BIO.BIO object.

    :param bio: M2Crypto.BIO.BIO object containing RSA key pair in PEM
                format.

    :param callback: A Python callable object that is invoked
                     to acquire a passphrase with which to unlock the
                     key.  The default is util.passphrase_callback.

    :return: M2Crypto.RSA.RSA object.
    """
    if bio.bio is None:
        raise RSAError("Uninitialized bio.bio")
    rsa = m2.rsa_read_key(bio.bio, callback)
    if rsa is None:
        rsa_error()
    return RSA(rsa, 1)


def load_key_string(
    string: Union[str, bytes],
    callback: Callable[..., Any] = util.passphrase_callback,
) -> RSA:
    """
    Load an RSA key pair from a string.

    :param string: String containing RSA key pair in PEM format.

    :param callback: A Python callable object that is invoked
                     to acquire a passphrase with which to unlock the
                     key. The default is util.passphrase_callback.

    :return: M2Crypto.RSA.RSA object.
    """
    bytes_to_load = string.encode("utf-8") if isinstance(string, str) else string
    bio = BIO.MemoryBuffer(bytes_to_load)
    return load_key_bio(bio, callback)


def load_pub_key(file: Union[str, bytes]) -> RSA_pub:
    """
    Load an RSA public key from file.

    :param file: Name of file containing RSA public key in PEM format.

    :return: M2Crypto.RSA.RSA_pub object.
    """
    with BIO.openfile(file) as bio:
        return load_pub_key_bio(bio)


def load_pub_key_bio(bio: BIO.BIO) -> RSA_pub:
    """
    Load an RSA public key from an M2Crypto.BIO.BIO object.

    :param bio: M2Crypto.BIO.BIO object containing RSA public key in PEM
                format.

    :return: M2Crypto.RSA.RSA_pub object.
    """
    if bio.bio is None:
        raise RSAError("Uninitialized bio.bio")
    rsa = m2.rsa_read_pub_key(bio.bio)
    if rsa is None:
        rsa_error()
    return RSA_pub(rsa, 1)


def new_pub_key(e_n: Tuple[bytes, bytes]) -> RSA_pub:
    """
    Instantiate an RSA_pub object from an (e, n) tuple.

    :param e: The RSA public exponent; it is a string in OpenSSL's MPINT
              format - 4-byte big-endian bit-count followed by the
              appropriate number of bits.

    :param n: The RSA composite of primes; it is a string in OpenSSL's
              MPINT format - 4-byte big-endian bit-count followed by the
              appropriate number of bits.

    :return: M2Crypto.RSA.RSA_pub object.
    """
    (e, n) = e_n
    rsa = m2.rsa_new()
    m2.rsa_set_en(rsa, e, n)
    return RSA_pub(rsa, 1)



"""
Make Twisted use M2Crypto for SSL

Copyright (c) 2004-2007 Open Source Applications Foundation.
All rights reserved.

FIXME THIS HAS NOT BEEN FINISHED. NEITHER PEP484 NOR PORT PYTHON3 HAS
BEEN FINISHED. THE FURTHER WORK WILL BE DONE WHEN THE STATUS OF TWISTED
IN THE PYTHON 3 (AND ASYNCIO) WORLD WILL BE CLEAR.
"""

__all__ = [
    "connectSSL",
    "connectTCP",
    "listenSSL",
    "listenTCP",
    "TLSProtocolWrapper",
]

import logging

from functools import partial

import twisted.internet.reactor
import twisted.protocols.policies as policies

from M2Crypto import BIO, X509, m2, util, types as C
from M2Crypto.SSL.Checker import Checker, SSLVerificationError
from M2Crypto.SSL import Context

from twisted.internet.interfaces import (
    ITLSTransport,
    IProtocolFactory,
    IProtocol,
    IReactorTCP,
    IReactorSSL,
)
from twisted.protocols.policies import ProtocolWrapper
from twisted.python.failure import Failure
from typing import Callable, Iterable, Optional, Union, Any  # noqa
from zope.interface import implementer  # type: ignore[import-untyped]

log = logging.getLogger(__name__)


def _alwaysSucceedsPostConnectionCheck(
    peerX509: Optional[X509.X509], expectedHost: str
) -> int:
    return 1


def connectSSL(
    host: str,
    port: int,
    factory: IProtocolFactory,
    contextFactory: Context,
    timeout: int = 30,
    bindAddress: Optional[str] = None,
    reactor: IReactorSSL = twisted.internet.reactor,  # type: ignore[assignment]
    postConnectionCheck: Checker = Checker(),
) -> object:
    """
    A convenience function to start an SSL/TLS connection using Twisted.

    See IReactorSSL interface in Twisted.
    """
    wrappingFactory = policies.WrappingFactory(factory)
    wrappingFactory.protocol = lambda factory, wrappedProtocol: TLSProtocolWrapper(  # type: ignore[assignment]
        factory,
        wrappedProtocol,
        startPassThrough=0,
        client=1,
        contextFactory=contextFactory,
        postConnectionCheck=postConnectionCheck,
    )
    return reactor.connectTCP(host, port, wrappingFactory, timeout, bindAddress)  # type: ignore[call-arg, arg-type]


def connectTCP(
    host: str,
    port: int,
    factory: IProtocolFactory,
    timeout: int = 30,
    bindAddress: Optional[util.AddrType] = None,
    reactor: IReactorTCP = twisted.internet.reactor,  # type: ignore[assignment]
    postConnectionCheck: Callable = Checker(),
) -> object:
    """
    A convenience function to start a TCP connection using Twisted.

    NOTE: You must call startTLS(ctx) to go into SSL/TLS mode.

    See IReactorTCP interface in Twisted.
    """
    wrappingFactory = policies.WrappingFactory(factory)
    wrappingFactory.protocol = lambda factory, wrappedProtocol: TLSProtocolWrapper(  # type: ignore[assignment]
        factory,
        wrappedProtocol,
        startPassThrough=1,
        client=1,
        contextFactory=None,
        postConnectionCheck=postConnectionCheck,
    )
    return reactor.connectTCP(host, port, wrappingFactory, timeout, bindAddress)  # type: ignore[call-arg, arg-type, misc]


def listenSSL(
    port: Union[int, str],
    factory: IProtocolFactory,
    contextFactory: Context,
    backlog: int = 5,
    interface: str = "",
    reactor: IReactorSSL = twisted.internet.reactor,  # type: ignore[assignment]
    postConnectionCheck: Callable = _alwaysSucceedsPostConnectionCheck,
) -> object:
    """
    A convenience function to listen for SSL/TLS connections using Twisted.

    See IReactorSSL interface in Twisted.
    """
    wrappingFactory = policies.WrappingFactory(factory)
    wrappingFactory.protocol = lambda factory, wrappedProtocol: TLSProtocolWrapper(  # type: ignore[assignment]
        factory,
        wrappedProtocol,
        startPassThrough=0,
        client=0,
        contextFactory=contextFactory,
        postConnectionCheck=postConnectionCheck,
    )
    return reactor.listenTCP(port, wrappingFactory, backlog, interface)  # type: ignore[call-arg, arg-type]


def listenTCP(
    port: Union[int, str],
    factory: IProtocolFactory,
    backlog: int = 5,
    interface: str = "",
    reactor: IReactorTCP = twisted.internet.reactor,  # type: ignore[assignment]
    postConnectionCheck: Optional[Callable] = None,
) -> object:
    """
    A convenience function to listen for TCP connections using Twisted.

    NOTE: You must call startTLS(ctx) to go into SSL/TLS mode.

    See IReactorTCP interface in Twisted.
    """
    wrappingFactory = policies.WrappingFactory(factory)
    wrappingFactory.protocol = lambda factory, wrappedProtocol: TLSProtocolWrapper(  # type: ignore[assignment]
        factory,
        wrappedProtocol,
        startPassThrough=1,
        client=0,
        contextFactory=None,
        postConnectionCheck=postConnectionCheck,
    )
    return reactor.listenTCP(port, wrappingFactory, backlog, interface)  # type: ignore[call-arg, arg-type, misc]


class _BioProxy(object):
    """
    The purpose of this class is to eliminate the __del__ method from
    TLSProtocolWrapper, and thus letting it be garbage collected.
    """

    def __init__(self, bio: BIO.BIO) -> None:
        self.bio = bio

    @staticmethod
    def m2_bio_free_all(bio: BIO.BIO):
        m2.bio_free_all(bio)  # type: ignore[arg-type]

    def _ptr(self) -> C.BIO:
        return self.bio  # type: ignore[return-value]

    def __del__(self) -> None:
        if self.bio is not None:
            # self.m2_bio_free_all(self.bio) # Mypy complains about types, better to let it go
            pass


class _SSLProxy(object):
    """
    The purpose of this class is to eliminate the __del__ method from
    TLSProtocolWrapper, and thus letting it be garbage collected.
    """

    def __init__(self, ssl: C.SSL) -> None:
        self.ssl = ssl

    @staticmethod
    def m2_ssl_free(ssl: Any):
        m2.ssl_free(ssl)  # type: ignore[arg-type]

    def _ptr(self) -> C.SSL:
        return self.ssl  # type: ignore[return-value]

    def __del__(self) -> None:
        if self.ssl is not None:
            # self.m2_ssl_free(self.ssl) # Mypy complains about types, better to let it go
            pass


@implementer(ITLSTransport)
class TLSProtocolWrapper(ProtocolWrapper):
    """
    A SSL/TLS protocol wrapper to be used with Twisted. Typically
    you would not use this class directly. Use connectTCP,
    connectSSL, listenTCP, listenSSL functions defined above,
    which will hook in this class.
    """

    ctx: Optional[Context] = None
    sslBio: Optional[_BioProxy] = None
    ssl: Optional[_SSLProxy] = None
    internalBio: Any = None
    networkBio: Optional[_BioProxy] = None
    data: bytes
    encrypted: bytes
    tlsStarted: int
    checked: int
    isClient: int
    helloDone: int
    postConnectionCheck: Callable

    def __init__(
        self,
        factory: policies.WrappingFactory,
        wrappedProtocol: IProtocol,
        startPassThrough: int,
        client: int,
        contextFactory: Optional[Context],
        postConnectionCheck: Optional[Callable],
    ) -> None:
        """
        :param factory:
        :param wrappedProtocol:
        :param startPassThrough:    If true we won't encrypt at all. Need to
                                    call startTLS() later to switch to SSL/TLS.
        :param client:              True if this should be a client protocol.
        :param contextFactory:      Factory that creates SSL.Context objects.
                                    The called function is getContext().
        :param postConnectionCheck: The post connection check callback that
                                    will be called just after connection has
                                    been established but before any real data
                                    has been exchanged. The first argument to
                                    this function is an X509 object, the second
                                    is the expected host name string.
        """
        # ProtocolWrapper.__init__(self, factory, wrappedProtocol)
        # XXX: Twisted 2.0 has a new addition where the wrappingFactory is
        #      set as the factory of the wrappedProtocol. This is an issue
        #      as the wrap should be transparent. What we want is
        #      the factory of the wrappedProtocol to be the wrappedFactory and
        #      not the outer wrappingFactory. This is how it was implemented in
        #      Twisted 1.3
        self.factory = factory
        self.wrappedProtocol = wrappedProtocol

        # wrappedProtocol == client/server instance
        # factory.wrappedFactory == client/server factory

        self.data = b""  # Clear text to encrypt and send
        self.encrypted = b""  # Encrypted data we need to decrypt and pass on
        self.tlsStarted = 0  # SSL/TLS mode or pass through
        self.checked = 0  # Post connection check done or not
        self.isClient = client
        self.helloDone = 0  # True when hello has been sent
        if postConnectionCheck is None:
            self.postConnectionCheck = _alwaysSucceedsPostConnectionCheck
        else:
            self.postConnectionCheck = postConnectionCheck

        if not startPassThrough:
            if contextFactory is not None:
                self.startTLS(contextFactory.getContext())  # type: ignore[attr-defined]

    def clear(self) -> None:
        """
        Clear this instance, after which it is ready for reuse.
        """
        if getattr(self, "tlsStarted", 0):
            self.sslBio = None
            self.ssl = None
            self.internalBio = None
            self.networkBio = None
        self.data = b""
        self.encrypted = b""
        self.tlsStarted = 0
        self.checked = 0
        self.isClient = 1
        self.helloDone = 0
        # We can reuse self.ctx and it will be deleted automatically
        # when this instance dies

    def startTLS(self, ctx: Context) -> None:
        """
        Start SSL/TLS. If this is not called, this instance just passes data
        through untouched.
        """
        # NOTE: This method signature must match the startTLS() method Twisted
        #       expects transports to have. This will be called automatically
        #       by Twisted in STARTTLS situations, for example with SMTP.
        if self.tlsStarted:
            raise Exception("TLS already started")

        self.ctx = ctx

        # Corrected bio_s_bio to bio_s_mem based on previous assumed stub error, but will ignore the call
        self.internalBio = m2.bio_new(m2.bio_s_mem())  # type: ignore[attr-defined]
        m2.bio_set_write_buf_size(self.internalBio, 0)
        self.networkBio = _BioProxy(m2.bio_new(m2.bio_s_mem()))  # type: ignore[arg-type, attr-defined]
        m2.bio_set_write_buf_size(self.networkBio._ptr(), 0)
        m2.bio_make_bio_pair(self.internalBio, self.networkBio._ptr())  # type: ignore[attr-defined]

        self.sslBio = _BioProxy(m2.bio_new(m2.bio_f_ssl()))  # type: ignore[arg-type, attr-defined]

        self.ssl = _SSLProxy(m2.ssl_new(self.ctx.ctx))

        if self.isClient:
            m2.ssl_set_connect_state(self.ssl._ptr())
        else:
            m2.ssl_set_accept_state(self.ssl._ptr())

        m2.ssl_set_bio(self.ssl._ptr(), self.internalBio, self.internalBio)
        m2.bio_set_ssl(self.sslBio._ptr(), self.ssl._ptr(), m2.bio_noclose)

        # Need this for writes that are larger than BIO pair buffers
        mode = m2.ssl_get_mode(self.ssl._ptr())  # type: ignore[attr-defined]
        m2.ssl_set_mode(  # type: ignore[attr-defined]
            self.ssl._ptr(),
            mode
            | m2.SSL_MODE_ENABLE_PARTIAL_WRITE  # type: ignore[attr-defined]
            | m2.SSL_MODE_ACCEPT_MOVING_WRITE_BUFFER,  # type: ignore[attr-defined]
        )

        self.tlsStarted = 1

    def write(self, data: bytes) -> None:
        if not self.tlsStarted:
            ProtocolWrapper.write(self, data)
            return

        try:
            encryptedData = self._encrypt(data)
            ProtocolWrapper.write(self, encryptedData)
            self.helloDone = 1
        except BIO.BIOError as e:
            # See http://www.openssl.org/docs/apps/verify.html#DIAGNOSTICS
            # for the error codes returned by SSL_get_verify_result.
            e.args = (
                m2.ssl_get_verify_result(self.ssl._ptr()),  # type: ignore[union-attr]
                e.args[0],
            )
            raise e

    def writeSequence(self, data: Iterable[bytes]) -> None:
        if not self.tlsStarted:
            ProtocolWrapper.writeSequence(self, b"".join(data))
            return

        self.write(b"".join(data))

    def loseConnection(self) -> None:
        # XXX Do we need to do m2.ssl_shutdown(self.ssl._ptr())?
        ProtocolWrapper.loseConnection(self)

    def connectionMade(self) -> None:
        ProtocolWrapper.connectionMade(self)
        if self.tlsStarted and self.isClient and not self.helloDone:
            self._clientHello()

    def dataReceived(self, data: bytes) -> None:
        if not self.tlsStarted:
            ProtocolWrapper.dataReceived(self, data)
            return

        self.encrypted += data

        try:
            while 1:
                decryptedData = self._decrypt()

                self._check()

                encryptedData = self._encrypt()
                ProtocolWrapper.write(self, encryptedData)

                ProtocolWrapper.dataReceived(self, decryptedData)

                if decryptedData == b"" and encryptedData == b"":
                    break
        except BIO.BIOError as e:
            # See http://www.openssl.org/docs/apps/verify.html#DIAGNOSTICS
            # for the error codes returned by SSL_get_verify_result.
            e.args = (
                m2.ssl_get_verify_result(self.ssl._ptr()),  # type: ignore[union-attr]
                e.args[0],
            )
            raise e

    def connectionLost(
        self, reason: Optional[Failure] = None
    ) -> None:  # Added = None to match base class
        self.clear()
        ProtocolWrapper.connectionLost(self, reason)

    def _check(self) -> None:
        if not self.checked and self.ssl is not None and m2.ssl_is_init_finished(self.ssl._ptr()):  # type: ignore[attr-defined]
            x509 = m2.ssl_get_peer_cert(self.ssl._ptr())
            if x509 is not None:
                x509 = X509.X509(x509, 1)  # type: ignore[assignment]

            if self.isClient:
                if self.transport is not None:
                    host = self.transport.addr[0]  # type: ignore[union-attr]
                else:
                    return  # Exit if transport is missing
            else:
                if self.transport is not None:
                    host = self.transport.getPeer().host  # type: ignore[union-attr, misc]
                else:
                    return  # Exit if transport is missing

            if not self.postConnectionCheck(x509, host):
                raise SSLVerificationError("post connection check")
            self.checked = 1

    def _clientHello(self) -> None:
        try:
            # We rely on OpenSSL implicitly starting with client hello
            # when we haven't yet established an SSL connection
            encryptedData = self._encrypt(clientHello=1)
            ProtocolWrapper.write(self, encryptedData)
            self.helloDone = 1
        except BIO.BIOError as e:
            # See http://www.openssl.org/docs/apps/verify.html#DIAGNOSTICS
            # for the error codes returned by SSL_get_verify_result.
            e.args = (
                m2.ssl_get_verify_result(self.ssl._ptr()),  # type: ignore[union-attr]
                e.args[0],
            )
            raise e

    # Optimizations to reduce attribute accesses

    @property
    def _get_wr_guar_ssl(self) -> Callable[[], int]:
        """Return max. length of data can be written to the BIO.

        Writes larger than this value will return a value from
        BIO_write() less than the amount requested or if the buffer is
        full request a retry.
        """
        return partial(m2.bio_ctrl_get_write_guarantee, self.sslBio._ptr())  # type: ignore[union-attr]

    @property
    def _get_wr_guar_net(self) -> Callable[[], int]:
        return partial(m2.bio_ctrl_get_write_guarantee, self.networkBio._ptr())  # type: ignore[union-attr]

    @property
    def _shoud_retry_ssl(self) -> Callable[[], int]:
        # BIO_should_retry() is true if the call that produced this
        # condition should then be retried at a later time.
        return partial(m2.bio_should_retry, self.sslBio._ptr())  # type: ignore[union-attr]

    @property
    def _shoud_retry_net(self) -> Callable[[], int]:
        return partial(m2.bio_should_retry, self.networkBio._ptr())  # type: ignore[union-attr]

    @property
    def _ctrl_pend_ssl(self) -> Callable[[], int]:
        # size_t BIO_ctrl_pending(BIO *b);
        # BIO_ctrl_pending() return the number of pending characters in
        # the BIOs read and write buffers.
        return partial(m2.bio_ctrl_pending, self.sslBio._ptr())  # type: ignore[union-attr]

    @property
    def _ctrl_pend_net(self) -> Callable[[], int]:
        return partial(m2.bio_ctrl_pending, self.networkBio._ptr())  # type: ignore[union-attr]

    @property
    def _write_ssl(self) -> Callable[[bytes], int]:
        # All these functions return either the amount of data
        # successfully read or written (if the return value is
        # positive) or that no data was successfully read or written
        # if the result is 0 or -1. If the return value is -2 then
        # the operation is not implemented in the specific BIO type.
        return partial(m2.bio_write, self.sslBio._ptr())  # type: ignore[union-attr]

    @property
    def _write_net(self) -> Callable[[bytes], int]:
        return partial(m2.bio_write, self.networkBio._ptr())  # type: ignore[union-attr]

    @property
    def _read_ssl(self) -> Callable[[int], Optional[bytes]]:
        return partial(m2.bio_read, self.sslBio._ptr())  # type: ignore[union-attr]

    @property
    def _read_net(self) -> Callable[[int], Optional[bytes]]:
        return partial(m2.bio_read, self.networkBio._ptr())  # type: ignore[union-attr]

    def _encrypt(self, data: bytes = b"", clientHello: int = 0) -> bytes:
        """
        :param data:
        :param clientHello:
        :return:
        """
        encryptedData = b""
        self.data += data

        while 1:
            if (self._get_wr_guar_ssl() > 0 and self.data != b"") or clientHello:
                r = self._write_ssl(self.data)
                if r <= 0:
                    if not self._shoud_retry_ssl():
                        raise IOError(
                            (
                                "Data left to be written to {}, "
                                + "but cannot retry SSL connection!"
                            ).format(self.sslBio)
                        )
                else:
                    assert self.checked
                    self.data = self.data[r:]

            pending = self._ctrl_pend_net()
            if pending:
                d = self._read_net(pending)
                if d is not None:  # This is strange, but d can be None
                    encryptedData += d
                else:
                    assert self._shoud_retry_net()
            else:
                break
        return encryptedData

    def _decrypt(self, data: bytes = b"") -> bytes:
        self.encrypted += data
        decryptedData = b""

        while 1:
            if self._get_wr_guar_ssl() > 0 and self.encrypted != b"":
                r = self._write_net(self.encrypted)
                if r <= 0:
                    if not self._shoud_retry_net():
                        raise IOError(
                            (
                                "Data left to be written to {}, "
                                + "but cannot retry SSL connection!"
                            ).format(self.networkBio)
                        )
                else:
                    self.encrypted = self.encrypted[r:]

            pending = self._ctrl_pend_ssl()
            if pending:
                d = self._read_ssl(pending)
                if d is not None:  # This is strange, but d can be None
                    decryptedData += d
                else:
                    assert self._shoud_retry_ssl()
            else:
                break

        return decryptedData



"""
M2Crypto enhancement to Python's urllib2 for handling
'https' url's.

Code from urllib2 is Copyright (c) 2001, 2002, 2003, 2004, 2005, 2006, 2007
Python Software Foundation; All Rights Reserved

Summary of changes:
 - Use an HTTPSProxyConnection if the request is going through a proxy.
 - Add the SSL context to the https connection when performing https_open.
 - Add the M2Crypto HTTPSHandler when building a default opener.
"""

import socket

from M2Crypto import SSL, httpslib

from urllib.parse import urldefrag, urlparse as url_parse
from urllib.request import AbstractHTTPHandler
from urllib.response import addinfourl
from typing import Optional, Type  # noqa

from urllib.request import *  # noqa other modules want to import
from urllib.error import *  # noqa other modules want to import


def _makefile(sock_like, mode, bufsize):
    """
    The original implementation of this function created an infinite
    recursion by incorrectly monkey-patching _decref_socketios.
    Removing the incorrect logic resolves the issue. The SocketIO
    object correctly calls close() on the wrapped sock_like object
    by default.
    """
    return socket.SocketIO(sock_like, mode)


class RefCountingSSLConnection(SSL.Connection):
    """A reference counting SSL connection.

    It can be wrapped into a socket._fileobject or socket.SocketIO instance.
    If the wrapping object is closed or subject to garbage collection,
    this SSL connection is only shut down if there are no more references,
    which were created by RefCountingSSLConnection.makefile, to it.
    """

    def __init__(self, *args, **kwargs):
        SSL.Connection.__init__(self, *args, **kwargs)
        # Start with one reference for the connection object itself.
        self._refs = 0
        self._closed = False

    def _decref_socketios(self):
        if self._refs > 0:
            self._refs -= 1
        if self._refs == 0 and not self._closed:
            # make sure we close the connection only once
            # (otherwise we end up with a bidirectional shutdown)
            self._closed = True
            super(RefCountingSSLConnection, self).close()

    def close(self):
        """
        Close the connection. This is idempotent.

        The original ref-counting logic has been bypassed for this direct
        call to provide a simpler, more robust shutdown path for the
        primary use case.
        """
        if not getattr(self, "_closed", False):
            self._closed = True
            # Directly close the parent connection without complex logic.
            super(RefCountingSSLConnection, self).close()

    def makefile(self, mode="rb", bufsize=-1):
        self._refs += 1
        return _makefile(self, mode, bufsize)


class HTTPSHandler(AbstractHTTPHandler):  # type: ignore [no-redef]
    def __init__(
        self,
        ssl_context: Optional[SSL.Context] = None,
        ssl_conn_cls: Type[SSL.Connection] = RefCountingSSLConnection,
    ):
        AbstractHTTPHandler.__init__(self)

        if ssl_context is not None:
            assert isinstance(ssl_context, SSL.Context), ssl_context
            self.ctx = ssl_context
        else:
            self.ctx = SSL.Context()
        self._ssl_conn_cls = ssl_conn_cls

    # Copied from urllib2, so we can set the ssl context.
    def https_open(self, req: Request) -> addinfourl:
        """Return an addinfourl object for the request, using http_class.

        http_class must implement the HTTPConnection API from httplib.
        The addinfourl return value is a file-like object.  It also
        has methods and attributes including:

            - info(): return a mimetools.Message object for the headers

            - geturl(): return the original request URL

            - code: HTTP status code
        """
        host = req.host
        if not host:
            raise URLError("no host given")

        # Our change: Check to see if we're using a proxy.
        # Then create an appropriate ssl-aware connection.
        full_url = req.get_full_url()
        target_host = url_parse(full_url)[1]

        # Explicitly type `h` to the base class to handle both branches.
        h: httpslib.HTTPSConnection

        if target_host != host:
            request_uri = urldefrag(full_url)[0]
            # Mypy gets confused by re-defined classes, so we ignore errors.
            h = httpslib.ProxyHTTPSConnection(  # type: ignore[call-arg]
                host=host,
                ssl_context=self.ctx,
                ssl_conn_cls=self._ssl_conn_cls,
            )
        else:
            request_uri = req.selector
            # Mypy gets confused by re-defined classes, so we ignore errors.
            h = httpslib.HTTPSConnection(  # type: ignore[call-arg]
                host=host,
                ssl_context=self.ctx,
                ssl_conn_cls=self._ssl_conn_cls,
            )

        # The parent class has this attribute, mypy is just confused.
        h.set_debuglevel(self._debuglevel)  # type: ignore[attr-defined]

        headers = dict(req.headers)
        headers.update(req.unredirected_hdrs)
        headers["Connection"] = "close"
        try:
            h.request(req.get_method(), request_uri, req.data, headers)
            r = h.getresponse()
        except (socket.error, SSL.SSLError) as err:
            h.close()  # Ensure cleanup on failure.
            raise URLError(err)

        # The HTTPResponse object 'r' is the file-like object we need.
        # The following lines monkey-patch 'r' to add attributes that older
        # versions of urllib expected.
        r.recv = r.read  # type: ignore[attr-defined]
        r.ssl = h.sock  # type: ignore[attr-defined]

        # Use the modern .headers attribute, not the deprecated .msg.
        resp = addinfourl(r, r.headers, req.get_full_url())
        resp.code = r.status
        resp.msg = r.reason  # type: ignore[attr-defined]

        # Attach the connection to the response to prevent premature GC.
        resp._connection = h  # type: ignore[attr-defined]

        # Hijack the close method to ensure the underlying SSL connection closes.
        the_connection_to_close = h.sock
        original_close = resp.close

        def new_close() -> None:
            try:
                original_close()
            finally:
                del resp._connection
                if the_connection_to_close:
                    the_connection_to_close.close()

        # Tell mypy to ignore the assignment to a method.
        resp.close = new_close  # type: ignore[method-assign]

        return resp

    https_request = AbstractHTTPHandler.do_request_


# Copied from urllib2 with modifications for ssl
def build_opener(  # type: ignore [no-redef]
    ssl_context: Optional[SSL.Context] = None, *handlers
) -> OpenerDirector:
    """Create an opener object from a list of handlers.

    The opener will use several default handlers, including support
    for HTTP and FTP.

    If any of the handlers passed as arguments are subclasses of the
    default handlers, the default handlers will not be used.
    """

    def isclass(obj):
        return isinstance(obj, type) or hasattr(obj, "__bases__")

    opener = OpenerDirector()
    default_classes = [
        ProxyHandler,
        UnknownHandler,
        HTTPHandler,
        HTTPDefaultErrorHandler,
        HTTPRedirectHandler,
        FTPHandler,
        FileHandler,
        HTTPErrorProcessor,
    ]
    skip = []
    for klass in default_classes:
        for check in handlers:
            if isclass(check):
                if issubclass(check, klass):
                    skip.append(klass)
            elif isinstance(check, klass):
                skip.append(klass)
    for klass in skip:
        default_classes.remove(klass)

    for klass in default_classes:
        opener.add_handler(klass())  # type: ignore[call-arg]

    # Add the HTTPS handler with ssl_context
    if HTTPSHandler not in skip:
        opener.add_handler(HTTPSHandler(ssl_context))  # type: ignore[arg-type]

    for h in handlers:
        if isclass(h):
            h = h()
        opener.add_handler(h)
    return opener



# mypy: allow-untyped-defs
"""
Fully populated mypy stub for M2Crypto's C bindings (_m2crypto)
Generated based on _m2crypto_wrap.c
"""

from typing import Any, Callable, List, Optional, Tuple, overload

# char * is str and it is for string constants only; NO R/W!

from M2Crypto import ASN1, BIO, types as C

# --- Constants ---

OPENSSL_VERSION_NUMBER: int
bio_close: int
bio_noclose: int
encrypt: int
decrypt: int
no_padding: int
pkcs1_padding: int
pkcs1_oaep_padding: int
sslv2_padding: int

# PKCS7 flags
PKCS7_BINARY: int
PKCS7_NOCERTS: int
PKCS7_NOCHAIN: int
PKCS7_NOINTERN: int
PKCS7_NOSIGS: int
PKCS7_NOVERIFY: int
PKCS7_TEXT: int
PKCS7_DETACHED: int
PKCS7_NOATTR: int
PKCS7_SIGNED: int
PKCS7_ENVELOPED: int
PKCS7_SIGNED_ENVELOPED: int
PKCS7_DATA: int
PKCS5_SALT_LEN: int

# SSL verify modes
SSL_VERIFY_NONE: int
SSL_VERIFY_PEER: int
SSL_VERIFY_FAIL_IF_NO_PEER_CERT: int
SSL_VERIFY_CLIENT_ONCE: int

# SSL options
SSL_OP_ALL: int
SSL_OP_NO_SSLv2: int
SSL_OP_NO_SSLv3: int
SSL_OP_NO_TLSv1: int

# SSL state flags
SSL_ST_CONNECT: int
SSL_ST_ACCEPT: int
SSL_ST_MASK: int
SSL_ST_INIT: int
SSL_ST_BEFORE: int
SSL_ST_OK: int
SSL_ST_RENEGOTIATE: int

# SSL callback flags
SSL_CB_LOOP: int
SSL_CB_EXIT: int
SSL_CB_READ: int
SSL_CB_WRITE: int
SSL_CB_ALERT: int
SSL_CB_READ_ALERT: int
SSL_CB_WRITE_ALERT: int
SSL_CB_ACCEPT_LOOP: int
SSL_CB_ACCEPT_EXIT: int
SSL_CB_CONNECT_LOOP: int
SSL_CB_CONNECT_EXIT: int
SSL_CB_HANDSHAKE_START: int
SSL_CB_HANDSHAKE_DONE: int

SSL_SENT_SHUTDOWN: int
SSL_RECEIVED_SHUTDOWN: int

# DH Generators
DH_GENERATOR_2: int
DH_GENERATOR_5: int

# EC constants
NID_secp112r1: int
NID_secp112r2: int
NID_secp128r1: int
NID_secp128r2: int
NID_secp160k1: int
NID_secp160r1: int
NID_secp160r2: int
NID_secp192k1: int
NID_secp224k1: int
NID_secp224r1: int
NID_secp256k1: int
NID_secp384r1: int
NID_secp521r1: int
NID_sect113r1: int
NID_sect113r2: int
NID_sect131r1: int
NID_sect131r2: int
NID_sect163k1: int
NID_sect163r1: int
NID_sect163r2: int
NID_sect193r1: int
NID_sect193r2: int
NID_sect233k1: int
NID_sect233r1: int
NID_sect239k1: int
NID_sect283k1: int
NID_sect283r1: int
NID_sect409k1: int
NID_sect409r1: int
NID_sect571k1: int
NID_sect571r1: int
NID_X9_62_prime192v1: int
NID_X9_62_prime192v2: int
NID_X9_62_prime192v3: int
NID_X9_62_prime239v1: int
NID_X9_62_prime239v2: int
NID_X9_62_prime239v3: int
NID_X9_62_prime256v1: int
NID_X9_62_c2pnb163v1: int
NID_X9_62_c2pnb163v2: int
NID_X9_62_c2pnb163v3: int
NID_X9_62_c2pnb176v1: int
NID_X9_62_c2tnb191v1: int
NID_X9_62_c2tnb191v2: int
NID_X9_62_c2tnb191v3: int
NID_X9_62_c2pnb208w1: int
NID_X9_62_c2tnb239v1: int
NID_X9_62_c2tnb239v2: int
NID_X9_62_c2tnb239v3: int
NID_X9_62_c2pnb272w1: int
NID_X9_62_c2pnb304w1: int
NID_X9_62_c2tnb359v1: int
NID_X9_62_c2pnb368w1: int
NID_X9_62_c2tnb431r1: int
NID_wap_wsg_idm_ecid_wtls1: int
NID_wap_wsg_idm_ecid_wtls3: int
NID_wap_wsg_idm_ecid_wtls4: int
NID_wap_wsg_idm_ecid_wtls5: int
NID_wap_wsg_idm_ecid_wtls6: int
NID_wap_wsg_idm_ecid_wtls7: int
NID_wap_wsg_idm_ecid_wtls8: int
NID_wap_wsg_idm_ecid_wtls9: int
NID_wap_wsg_idm_ecid_wtls10: int
NID_wap_wsg_idm_ecid_wtls11: int
NID_wap_wsg_idm_ecid_wtls12: int
NID_ipsec3: int
NID_ipsec4: int

# X509 constants
NID_commonName: int
NID_countryName: int
NID_localityName: int
NID_stateOrProvinceName: int
NID_organizationName: int
NID_organizationalUnitName: int
NID_serialNumber: int
NID_surname: int
NID_givenName: int
NID_pkcs9_emailAddress: int

X509_V_OK: int
X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT: int
X509_V_ERR_UNABLE_TO_GET_CRL: int
X509_V_ERR_UNABLE_TO_DECRYPT_CERT_SIGNATURE: int
X509_V_ERR_UNABLE_TO_DECRYPT_CRL_SIGNATURE: int
X509_V_ERR_UNABLE_TO_DECODE_ISSUER_PUBLIC_KEY: int
X509_V_ERR_CERT_SIGNATURE_FAILURE: int
X509_V_ERR_CRL_SIGNATURE_FAILURE: int
X509_V_ERR_CERT_NOT_YET_VALID: int
X509_V_ERR_CERT_HAS_EXPIRED: int
X509_V_ERR_CRL_NOT_YET_VALID: int
X509_V_ERR_CRL_HAS_EXPIRED: int
X509_V_ERR_ERROR_IN_CERT_NOT_BEFORE_FIELD: int
X509_V_ERR_ERROR_IN_CERT_NOT_AFTER_FIELD: int
X509_V_ERR_ERROR_IN_CRL_LAST_UPDATE_FIELD: int
X509_V_ERR_ERROR_IN_CRL_NEXT_UPDATE_FIELD: int
X509_V_ERR_OUT_OF_MEM: int
X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT: int
X509_V_ERR_SELF_SIGNED_CERT_IN_CHAIN: int
X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY: int
X509_V_ERR_UNABLE_TO_VERIFY_LEAF_SIGNATURE: int
X509_V_ERR_CERT_CHAIN_TOO_LONG: int
X509_V_ERR_CERT_REVOKED: int
X509_V_ERR_INVALID_CA: int
X509_V_ERR_PATH_LENGTH_EXCEEDED: int
X509_V_ERR_INVALID_PURPOSE: int
X509_V_ERR_CERT_UNTRUSTED: int
X509_V_ERR_CERT_REJECTED: int
X509_V_ERR_APPLICATION_VERIFICATION: int
VERIFY_ALLOW_PROXY_CERTS: int
VERIFY_CB_ISSUER_CHECK: int
VERIFY_CHECK_SS_SIGNATURE: int
VERIFY_CRL_CHECK: int
VERIFY_CRL_CHECK_ALL: int
VERIFY_CRL_CHECK_CHAIN: int
VERIFY_CRL_CHECK_LEAF: int
VERIFY_EXPLICIT_POLICY: int
VERIFY_EXTENDED_CRL_SUPPORT: int
VERIFY_IGNORE_CRITICAL: int
VERIFY_INHIBIT_ANY: int
VERIFY_INHIBIT_MAP: int
VERIFY_NO_ALT_CHAINS: int
VERIFY_NO_CHECK_TIME: int
VERIFY_NOTIFY_POLICY: int
VERIFY_PARTIAL_CHAIN: int
VERIFY_POLICY_CHECK: int
VERIFY_TRUSTED_FIRST: int
VERIFY_USE_DELTAS: int
VERIFY_X509_STRICT: int
XN_FLAG_COMPAT: int
XN_FLAG_SEP_COMMA_PLUS: int
XN_FLAG_SEP_CPLUS_SPC: int
XN_FLAG_SEP_MULTILINE: int
XN_FLAG_DN_REV: int
XN_FLAG_FN_LN: int
XN_FLAG_SPC_EQ: int
XN_FLAG_DUMP_UNKNOWN_FIELDS: int
XN_FLAG_FN_ALIGN: int
XN_FLAG_ONELINE: int
XN_FLAG_MULTILINE: int
XN_FLAG_RFC2253: int
RSA_3: int
RSA_F4: int

# --- Library Initialization and Utility Functions ---

def lib_init() -> None: ...
def time_t_bits() -> int: ...
def threading_init() -> None: ...
def threading_cleanup() -> None: ...

# --- Error Handling ---

def err_print_errors(bio: C.BIO) -> None: ...
def err_clear_error() -> None: ...
def err_get_error() -> int: ...
def err_peek_error() -> int: ...
def err_lib_error_string(err: int) -> str: ...
def err_func_error_string(err: int) -> str: ...
def err_reason_error_string(err: int) -> str: ...
def x509_get_verify_error(err: int) -> str: ...

# --- ASN.1 Functions ---

def asn1_integer_new() -> C.ASN1_Integer: ...
def asn1_integer_cmp(a1: C.ASN1_Integer, a2: C.ASN1_Integer) -> int: ...
def asn1_integer_free(a: C.ASN1_Integer) -> None: ...
def asn1_integer_get(a: C.ASN1_Integer) -> Optional[int]: ...
def asn1_integer_set(a: C.ASN1_Integer, v: int) -> int: ...
def asn1_string_new() -> C.ASN1_String: ...
def asn1_string_free(a: C.ASN1_String) -> None: ...
def asn1_string_print(out: C.BIO, s: C.ASN1_String) -> int: ...
def asn1_string_print_ex(out: C.BIO, s: C.ASN1_String, flags: int) -> int: ...
def asn1_string_set(s: C.ASN1_String, data: bytes) -> int: ...
def asn1_time_free(a: C.ASN1_Time) -> None: ...
def asn1_time_new() -> C.ASN1_Time: ...
def asn1_time_print(out: C.BIO, s: C.ASN1_Time) -> int: ...
def asn1_time_set(a: C.ASN1_Time, t: int) -> C.ASN1_Time: ...
def asn1_time_set_string(a: C.ASN1_Time, s: str) -> int: ...
def asn1_time_type_check(a: C.ASN1_Time) -> int: ...
def asn1_object_free(a: C.ASN1_Object) -> None: ...

# --- BIO (Basic I/O) Functions ---
def bio_ctrl_get_write_guarantee(b: C.BIO) -> int: ...
def bio_ctrl_pending(b: C.BIO) -> int: ...
def bio_ctrl_wpending(b: C.BIO) -> int: ...
def bio_do_handshake(b: C.BIO) -> int: ...
def bio_eof(b: C.BIO) -> int: ...
def bio_f_buffer() -> C.BIO_METHOD: ...
def bio_f_cipher() -> C.BIO_METHOD: ...
def bio_flush(b: C.BIO) -> int: ...
def bio_free_all(b: C.BIO) -> None: ...
def bio_free(b: C.BIO) -> int: ...
def bio_f_ssl() -> C.BIO_METHOD: ...
def bio_get_fd(b: C.BIO) -> int: ...
def bio_gets(b: C.BIO, size: int) -> bytes: ...
def bio_init(err: Any) -> None: ...
def bio_new_file(filename: str, mode: str) -> C.BIO: ...
def bio_new(method: C.BIO_METHOD) -> C.BIO: ...
def bio_new_pyfile(pyfile: Any, close_on_del: int) -> C.BIO: ...
def bio_new_socket(sock: int, close_flag: int) -> C.BIO: ...
def bio_pop(b: C.BIO) -> C.BIO: ...
def bio_push(b: C.BIO, append: C.BIO) -> C.BIO: ...
def bio_read(b: C.BIO, size: int) -> bytes: ...
def bio_reset(b: C.BIO) -> int: ...
def bio_seek(b: C.BIO, offset: int) -> int: ...
def bio_set_cipher(
    b: C.BIO, c: C.EVP_CIPHER, key: bytes, iv: bytes, op: int
) -> None: ...
def bio_set_flags(b: C.BIO, flags: int) -> None: ...
def bio_set_mem_eof_return(b: C.BIO, v: int) -> int: ...
def bio_set_ssl(b: C.BIO, ssl: C.SSL, flag: int) -> int: ...
def bio_set_write_buf_size(b: C.BIO, size: int) -> int: ...
def bio_should_read(b: C.BIO) -> int: ...
def bio_should_retry(b: C.BIO) -> int: ...
def bio_should_write(b: C.BIO) -> int: ...
def bio_s_mem() -> C.BIO_METHOD: ...
def bio_s_socket() -> C.BIO_METHOD: ...
def bio_write(b: C.BIO, data: bytes) -> int: ...

# --- Big Number (BIGNUM) Functions ---
def bn_to_mpi(bn: C.BIGNUM) -> bytes: ...
def mpi_to_bn(s: bytes) -> C.BIGNUM: ...
def bn_to_bin(bn: C.BIGNUM) -> bytes: ...
def bin_to_bn(s: bytes) -> C.BIGNUM: ...
def bn_to_hex(bn: C.BIGNUM) -> str: ...
def hex_to_bn(s: str) -> C.BIGNUM: ...
def dec_to_bn(s: str) -> C.BIGNUM: ...
def bn_rand(bits: int, top: int, bottom: int) -> int: ...
def bn_rand_range(range_bn: int) -> int: ...

# --- DH (Diffie-Hellman) Functions ---
def dh_compute_key(dh: C.DH, pubkey: bytes) -> bytes: ...
def dh_free(dh: C.DH) -> None: ...
def dh_generate_key(dh: C.DH) -> int: ...
def dh_generate_parameters(
    plen: int, g: int, pyfunc: Callable[[int, int], int]
) -> C.DH: ...
def dh_check(dh: C.DH) -> int: ...
def dh_init(err: Any) -> None: ...
def dh_new() -> C.DH: ...
def dhparams_print(bio: C.BIO, dh: C.DH) -> int: ...
def dh_read_parameters(bio: C.BIO) -> Optional[C.DH]: ...
def dh_set_pg(dh: C.DH, p: bytes, g: bytes) -> int: ...
def dh_size(dh: C.DH) -> int: ...
def dh_type_check(dh: C.DH) -> int: ...

# --- DSA Functions ---
def dsa_free(dsa: C.DSA) -> None: ...
def dsa_generate_key(dsa: C.DSA) -> int: ...
def dsa_generate_parameters(bits: int, pyfunc: Callable[[int, int], int]) -> C.DSA: ...
def dsa_gen_key(dsa: C.DSA) -> int: ...
def dsa_check_pub_key(dsa: C.DSA) -> int: ...
def dsa_init(err: Any) -> None: ...
def dsa_keylen(dsa: C.DSA) -> int: ...
def dsa_new() -> C.DSA: ...
def dsa_read_key(
    bio: C.BIO, pyfunc: Optional[Callable[[int], bytes]] = None
) -> Optional[C.DSA]: ...
def dsa_read_params(
    bio: C.BIO, pyfunc: Optional[Callable[[int], bytes]] = None
) -> Optional[C.DSA]: ...
def dsa_read_pub_key(
    bio: C.BIO, pyfunc: Optional[Callable[[int], bytes]] = None
) -> Optional[C.DSA]: ...
def dsa_set_pqg(dsa: C.DSA, p: bytes, q: bytes, g: bytes) -> int: ...
def dsa_set_pub(dsa: C.DSA, pub_key: bytes) -> int: ...
def dsa_sign_asn1(dsa: C.DSA, data: bytes) -> bytes: ...
def dsa_sign(dsa: C.DSA, data: bytes) -> Tuple[bytes, bytes]: ...
def dsa_type_check(dsa: C.DSA) -> int: ...
def dsa_verify_asn1(dsa: C.DSA, data: bytes, sig: bytes) -> int: ...
def dsa_verify(dsa: C.DSA, data: bytes, r: bytes, s: bytes) -> int: ...
def dsa_write_key_bio(
    dsa: C.DSA,
    bio: C.BIO,
    cipher: Optional[C.EVP_CIPHER],
    pyfunc: Callable[[int], bytes],
) -> int: ...
def dsa_write_key_bio_no_cipher(
    dsa: C.DSA, bio: C.BIO, pyfunc: Callable[[int], bytes]
) -> int: ...
def dsa_write_params_bio(dsa: C.DSA, bio: C.BIO) -> int: ...
def dsa_write_pub_key_bio(dsa: C.DSA, bio: C.BIO) -> int: ...

# --- EC (Elliptic Curve) Functions ---
def ecdh_compute_key(keypairA: C.EC_KEY, pubkeyB: C.EC_KEY) -> bytes: ...
def ecdsa_sign_asn1(key: C.EC_KEY, data: bytes) -> bytes: ...
def ecdsa_sign(key: C.EC_KEY, data: bytes) -> Tuple[bytes, bytes]: ...
def ecdsa_verify_asn1(key: C.EC_KEY, data: bytes, sig: bytes) -> int: ...
def ecdsa_verify(key: C.EC_KEY, data: bytes, r: bytes, s: bytes) -> int: ...
def ec_get_builtin_curves() -> List[Any]: ...
def ec_init(err: Any) -> None: ...
def ec_key_free(key: C.EC_KEY) -> None: ...
def ec_key_from_pubkey_der(pubkey: bytes) -> C.EC_KEY: ...
def ec_key_from_pubkey_params(nid: int, pubkey: bytes) -> C.EC_KEY: ...
def ec_key_gen_key(key: C.EC_KEY) -> int: ...
def ec_key_get_public_der(key: C.EC_KEY) -> bytes: ...
def ec_key_get_public_key(key: C.EC_KEY) -> bytes: ...
def ec_key_check_key(key: C.EC_KEY) -> int: ...
def ec_key_keylen(key: C.EC_KEY) -> int: ...
def ec_key_new_by_curve_name(nid: int) -> C.EC_KEY: ...
def ec_key_new() -> C.EC_KEY: ...
def ec_key_read_bio(
    bio: C.BIO, pyfunc: Optional[Callable[[int], bytes]] = None
) -> Optional[C.EC_KEY]: ...
def ec_key_read_pubkey(bio: C.BIO) -> Optional[C.EC_KEY]: ...
def ec_key_type_check(key: C.EC_KEY) -> int: ...
def ec_key_write_bio(
    key: C.EC_KEY,
    bio: C.BIO,
    cipher: Optional[C.EVP_CIPHER],
    pyfunc: Callable[[int], bytes],
) -> int: ...
def ec_key_write_bio_no_cipher(
    key: C.EC_KEY, bio: C.BIO, pyfunc: Callable[[int], bytes]
) -> int: ...
def ec_key_write_pubkey(key: C.EC_KEY, bio: C.BIO) -> int: ...

# --- EVP (High-level Crypto) Functions ---
def aes_128_cbc() -> C.EVP_CIPHER: ...
def aes_192_cbc() -> C.EVP_CIPHER: ...
def aes_256_cbc() -> C.EVP_CIPHER: ...
def bf_cbc() -> C.EVP_CIPHER: ...
def bytes_to_key(
    cipher: C.EVP_CIPHER,
    md: C.EVP_MD,
    data: bytes,
    salt: Optional[bytes],
    iv: Any,
    iter: int,
) -> bytes: ...
def cast5_cbc() -> C.EVP_CIPHER: ...
def cipher_ctx_free(ctx: C.EVP_CIPHER_CTX) -> None: ...
def cipher_ctx_new() -> C.EVP_CIPHER_CTX: ...
def cipher_final(ctx: C.EVP_CIPHER_CTX) -> bytes: ...
def cipher_init(
    ctx: C.EVP_CIPHER_CTX,
    cipher: C.EVP_CIPHER,
    key: Optional[bytes],
    iv: Optional[bytes],
    op: int,
) -> None: ...
def cipher_set_padding(ctx: C.EVP_CIPHER_CTX, padding: int) -> int: ...
def cipher_update(ctx: C.EVP_CIPHER_CTX, data: bytes) -> bytes: ...
def des_cbc() -> C.EVP_CIPHER: ...
def des_ecb() -> C.EVP_CIPHER: ...
def des_ede_cbc() -> C.EVP_CIPHER: ...
def des_ede3_cbc() -> C.EVP_CIPHER: ...
def digest_final(ctx: C.EVP_MD_CTX) -> bytes: ...
def digest_init(ctx: C.EVP_MD_CTX, md: C.EVP_MD) -> int: ...
def digest_sign(ctx: C.EVP_MD_CTX, msg: bytes) -> bytes: ...
def digest_sign_final(ctx: C.EVP_MD_CTX) -> bytes: ...
@overload
def digest_sign_init(ctx: C.EVP_MD_CTX, pkey: C.EVP_PKEY) -> int: ...
@overload
def digest_sign_init(
    ctx: C.EVP_MD_CTX,
    pctx: Optional[C.EVP_PKEY_CTX],
    md: C.EVP_MD,
    e: Optional[C.ENGINE],
    pkey: C.EVP_PKEY,
) -> int: ...
def digest_sign_update(ctx: C.EVP_MD_CTX, data: bytes) -> None: ...
def digest_update(ctx: C.EVP_MD_CTX, data: bytes) -> int: ...
def digest_verify(ctx: C.EVP_MD_CTX, sig: bytes, msg: bytes) -> int: ...
def digest_verify_final(ctx: C.EVP_MD_CTX, sig: bytes) -> int: ...
@overload
def digest_verify_init(ctx: C.EVP_MD_CTX, pkey: C.EVP_PKEY) -> int: ...
@overload
def digest_verify_init(
    ctx: C.EVP_MD_CTX,
    pctx: Optional[C.EVP_PKEY_CTX],
    md: C.EVP_MD,
    e: Optional[C.ENGINE],
    pkey: C.EVP_PKEY,
) -> int: ...
def digest_verify_update(ctx: C.EVP_MD_CTX, data: bytes) -> int: ...
def evp_init(err: Any) -> None: ...
def get_digestbyname(name: str) -> Optional[C.EVP_MD]: ...
def hmac_ctx_free(ctx: C.HMAC_CTX) -> None: ...
def hmac_ctx_new() -> C.HMAC_CTX: ...
def hmac_final(ctx: C.HMAC_CTX) -> bytes: ...
def hmac_init(ctx: C.HMAC_CTX, key: bytes, md: C.EVP_MD) -> None: ...
def hmac(key: bytes, data: bytes, md: C.EVP_MD) -> bytes: ...
def hmac_update(ctx: C.HMAC_CTX, data: bytes) -> None: ...
def md_ctx_free(ctx: C.EVP_MD_CTX) -> None: ...
def md_ctx_new() -> C.EVP_MD_CTX: ...
def md5() -> C.EVP_MD: ...
def pkcs5_pbkdf2_hmac_sha1(
    password: bytes, salt: bytes, iterations: int, key_len: int
) -> bytes: ...
def pkey_as_der(pkey: C.EVP_PKEY) -> bytes: ...
def pkey_assign_ec(pkey: C.EVP_PKEY, ec: C.EC_KEY) -> int: ...
def pkey_assign_rsa(pkey: C.EVP_PKEY, rsa: C.RSA) -> int: ...
def pkey_free(pkey: C.EVP_PKEY) -> None: ...
def pkey_get_modulus(pkey: C.EVP_PKEY) -> bytes: ...
def pkey_get_raw_pub_key_sha1(pkey: C.EVP_PKEY) -> bytes: ...
def pkey_get1_ec(pkey: C.EVP_PKEY) -> Optional[C.EC_KEY]: ...
def pkey_get1_rsa(pkey: C.EVP_PKEY) -> Optional[C.RSA]: ...
def pkey_new() -> C.EVP_PKEY: ...
def pkey_read_pem(bio: C.BIO, pyfunc: Callable[[], bytes]) -> C.EVP_PKEY: ...
def pkey_read_pem_pubkey(bio: C.BIO, pyfunc: Callable[[], bytes]) -> C.EVP_PKEY: ...
def pkey_set1_ec(pkey: C.EVP_PKEY, ec: C.EC_KEY) -> int: ...
def pkey_set1_rsa(pkey: C.EVP_PKEY, rsa: C.RSA) -> int: ...
def pkey_size(pkey: C.EVP_PKEY) -> int: ...
def pkey_write_pem_no_cipher(
    pkey: C.EVP_PKEY, bio: C.BIO, pyfunc: Callable[[], bytes]
) -> int: ...
def pkey_write_pem(
    pkey: C.EVP_PKEY, bio: C.BIO, cipher: C.EVP_CIPHER, pyfunc: Callable[[], bytes]
) -> int: ...
def rc2_40_cbc() -> C.EVP_CIPHER: ...
def rc4() -> C.EVP_CIPHER: ...
def ripemd160() -> C.EVP_MD: ...
def sha1() -> C.EVP_MD: ...
def sha224() -> C.EVP_MD: ...
def sha256() -> C.EVP_MD: ...
def sha384() -> C.EVP_MD: ...
def sha512() -> C.EVP_MD: ...
def sign_final(ctx: C.EVP_MD_CTX, pkey: C.EVP_PKEY) -> bytes: ...
def sign_init(ctx: C.EVP_MD_CTX, md: C.EVP_MD) -> int: ...
def sign_update(ctx: C.EVP_MD_CTX, data: bytes) -> None: ...
def verify_final(ctx: C.EVP_MD_CTX, sig: bytes, pkey: C.EVP_PKEY) -> int: ...
def verify_init(ctx: C.EVP_MD_CTX, md: C.EVP_MD) -> int: ...
def verify_update(ctx: C.EVP_MD_CTX, data: bytes) -> int: ...

# --- Misc Utility Functions ---
def util_init(err: Any) -> None: ...
def util_hex_to_string(data: bytes) -> str: ...
def util_string_to_hex(data: str) -> bytes: ...
def obj_obj2txt(obj: C.ASN1_Object, no_name: int) -> str: ...
def obj_ln2nid(ln: str) -> int: ...
def obj_nid2sn(nid: int) -> str: ...
def obj_sn2nid(sn: str) -> int: ...
def obj_txt2nid(txt: str) -> int: ...

# --- RSA ---
def rsa_init(err: Any) -> None: ...
def rsa_new() -> C.RSA: ...
def rsa_generate_key(
    bits: int, e: int, callback: Callable[[int, int], int]
) -> C.RSA: ...
def rsa_free(rsa: C.RSA) -> None: ...
def rsa_get_e(rsa: C.RSA) -> bytes: ...
def rsa_get_ex_data(rsa: C.RSA, index: int) -> Any: ...
def rsa_get_n(rsa: C.RSA) -> bytes: ...
def rsa_check_key(rsa: C.RSA) -> int: ...
def rsa_check_pub_key(rsa: C.RSA) -> int: ...
def rsa_padding_add_pkcs1_pss(
    rsa: C.RSA, digest: bytes, hash_obj: C.EVP_MD, salt_length: int
) -> bytes: ...
def rsa_private_decrypt(rsa: C.RSA, data: bytes, padding: int) -> Optional[bytes]: ...
def rsa_private_encrypt(rsa: C.RSA, data: bytes, padding: int) -> bytes: ...
def rsa_public_decrypt(rsa: C.RSA, data: bytes, padding: int) -> Optional[bytes]: ...
def rsa_public_encrypt(rsa: C.RSA, data: bytes, padding: int) -> bytes: ...
def rsa_read_key(bio: C.BIO, pyfunc: Callable[[int], bytes]) -> C.RSA: ...
def rsa_read_pub_key(bio: C.BIO) -> C.RSA: ...
def rsa_set_en(rsa: C.RSA, e: bytes, n: bytes) -> None: ...
def rsa_set_ex_data(rsa: C.RSA, index: int, data: Any) -> int: ...
def rsa_sign(rsa: C.RSA, digest: bytes, method_type: int) -> bytes: ...
def rsa_size(rsa: C.RSA) -> int: ...
def rsa_type_check(rsa: C.RSA) -> int: ...
def rsa_write_key(
    rsa: C.RSA, bio: C.BIO, cipher: C.EVP_CIPHER, pyfunc: Callable[[int], bytes]
) -> int: ...
def rsa_write_key_no_cipher(
    rsa: C.RSA, bio: C.BIO, pyfunc: Callable[[int], bytes]
) -> int: ...
def rsa_verify_pkcs1_pss(
    rsa: C.RSA, digest: bytes, signature: bytes, hash_obj: C.EVP_MD, salt_length: int
) -> int: ...
def rsa_verify(
    rsa: C.RSA, digest: bytes, signature: bytes, method_type: int
) -> int: ...
def rsa_write_key_der(rsa: C.RSA, bio: C.BIO) -> int: ...
def rsa_write_pub_key(rsa: C.RSA, bio: C.BIO) -> int: ...

# --- RC4 ---
def rc4_new() -> C.RC4_KEY: ...
def rc4_free(key: C.RC4_KEY) -> None: ...
def rc4_set_key(key: C.RC4_KEY, data: bytes) -> None: ...
def rc4_update(key: C.RC4_KEY, data: bytes) -> bytes: ...

# --- SSL / TLS ---
def i2d_ssl_session(bio: C.BIO, sess: C.SSL_SESSION) -> None: ...
def sk_ssl_cipher_num(stack: C.STACK_OF_SSL_CIPHER) -> int: ...
def sk_ssl_cipher_value(stack: C.STACK_OF_SSL_CIPHER, idx: int) -> C.SSL_CIPHER: ...
def ssl_accept(ssl: C.SSL, timeout: float = -1.0) -> int: ...
def ssl_cipher_get_bits(c: C.SSL_CIPHER) -> int: ...
def ssl_cipher_get_name(c: C.SSL_CIPHER) -> str: ...
def ssl_cipher_get_version(c: C.SSL_CIPHER) -> str: ...
def ssl_clear(ssl: C.SSL) -> int: ...
def ssl_connect(ssl: C.SSL, timeout: float = -1.0) -> int: ...
def ssl_ctx_add_session(ctx: C.SSL_CTX, sess: C.SSL_SESSION) -> int: ...
def ssl_ctx_free(ctx: C.SSL_CTX) -> None: ...
def ssl_ctx_get_cert_store(ctx: C.SSL_CTX) -> C.X509_STORE: ...
def ssl_ctx_get_session_cache_mode(ctx: C.SSL_CTX) -> int: ...
def ssl_ctx_get_session_timeout(ctx: C.SSL_CTX) -> int: ...
def ssl_ctx_get_verify_depth(ctx: C.SSL_CTX) -> int: ...
def ssl_ctx_get_verify_mode(ctx: C.SSL_CTX) -> int: ...
def ssl_ctx_check_privkey(ctx: C.SSL_CTX) -> int: ...
def ssl_ctx_load_verify_locations(
    ctx: C.SSL_CTX, cafile: Optional[str], capath: Optional[str]
) -> int: ...
def ssl_ctx_new(method: C.SSL_METHOD) -> C.SSL_CTX: ...
def ssl_ctx_passphrase_callback(
    ctx: C.SSL_CTX, pyfunc: Callable[[], bytes]
) -> None: ...
def ssl_ctx_remove_session(ctx: C.SSL_CTX, sess: C.SSL_SESSION) -> int: ...
def ssl_ctx_set_cache_size(ctx: C.SSL_CTX, size: int) -> int: ...
def ssl_ctx_set_cipher_list(ctx: C.SSL_CTX, ciphers: str) -> int: ...
def ssl_ctx_set_client_CA_list_from_file(ctx: C.SSL_CTX, ca_file: str) -> None: ...
def ssl_ctx_set_default_verify_paths(ctx: C.SSL_CTX) -> int: ...
def ssl_ctx_set_info_callback(ctx: C.SSL_CTX, callback: Callable) -> None: ...
def ssl_ctx_set_options(ctx: C.SSL_CTX, op: int) -> int: ...
def ssl_ctx_set_session_cache_mode(ctx: C.SSL_CTX, mode: int) -> int: ...
def ssl_ctx_set_session_id_context(ctx: C.SSL_CTX, sid_ctx: bytes) -> int: ...
def ssl_ctx_set_session_timeout(ctx: C.SSL_CTX, t: int) -> int: ...
def ssl_ctx_set_tmp_dh(ctx: C.SSL_CTX, dh: C.DH) -> int: ...
def ssl_ctx_set_tmp_dh_callback(ctx: C.SSL_CTX, callback: Callable) -> None: ...
def ssl_ctx_set_tmp_rsa(ctx: C.SSL_CTX, rsa: C.RSA) -> int: ...
def ssl_ctx_set_tmp_rsa_callback(ctx: C.SSL_CTX, callback: Callable) -> None: ...
def ssl_ctx_set_verify(
    ctx: C.SSL_CTX, mode: int, callback: Callable[[int, C.X509_STORE_CTX], int]
) -> None: ...
def ssl_ctx_set_verify_default(ctx: C.SSL_CTX, mode: int) -> None: ...
def ssl_ctx_set_verify_depth(ctx: C.SSL_CTX, depth: int) -> None: ...
def ssl_ctx_use_cert(ctx: C.SSL_CTX, cert_file: str) -> int: ...
def ssl_ctx_use_cert_chain(ctx: C.SSL_CTX, chain_file: str) -> int: ...
def ssl_ctx_use_pkey_privkey(ctx: C.SSL_CTX, pkey: C.EVP_PKEY) -> int: ...
def ssl_ctx_use_privkey(ctx: C.SSL_CTX, key_file: str) -> int: ...
def ssl_free(ssl: C.SSL) -> None: ...
def ssl_get_cipher_list(ssl: C.SSL, n: int) -> str: ...
def ssl_get_ciphers(ssl: C.SSL) -> C.STACK_OF_SSL_CIPHER: ...
def ssl_get_current_cipher(ssl: C.SSL) -> Optional[C.SSL_CIPHER]: ...
def ssl_get_default_session_timeout(ssl: C.SSL) -> int: ...
def ssl_get_error(ssl: C.SSL, ret: int) -> int: ...
def ssl_get_peer_cert_chain(ssl: C.SSL) -> Optional[C.STACK_OF_X509]: ...
def ssl_get_session(ssl: C.SSL) -> Optional[C.SSL_SESSION]: ...
def ssl_get_shutdown(ssl: C.SSL) -> int: ...
def ssl_get_verify_mode(ssl: C.SSL) -> int: ...
def ssl_get_verify_depth(ssl: C.SSL) -> int: ...
def ssl_get_peer_cert(ssl: C.SSL) -> Optional[C.X509]: ...
def ssl_get_ssl_ctx(ssl: C.SSL) -> C.SSL_CTX: ...
def ssl_get_state(ssl: C.SSL) -> str: ...
def ssl_get_state_v(ssl: C.SSL) -> str: ...
def ssl_get_verify_result(ssl: C.SSL) -> int: ...
def ssl_get_version(ssl: C.SSL) -> str: ...
def ssl_get_alert_type_v(err: int) -> str: ...
def ssl_get_alert_desc_v(err: int) -> str: ...
def ssl_init(ssl_err: Any, ssl_timeout_err: Any) -> None: ...
def ssl_is_init_finished(ssl: C.SSL) -> int: ...
def ssl_new(ctx: C.SSL_CTX) -> C.SSL: ...
def ssl_pending(ssl: C.SSL) -> int: ...
def ssl_read_nbio(ssl: C.SSL, num: int) -> Optional[bytes]: ...
def ssl_read(ssl: C.SSL, num: int, timeout: float = -1.0) -> Optional[bytes]: ...
def ssl_renegotiate(ssl: C.SSL) -> int: ...
def ssl_session_free(sess: C.SSL_SESSION) -> None: ...
def ssl_session_get_timeout(sess: C.SSL_SESSION) -> int: ...
def ssl_session_get_time(sess: C.SSL_SESSION) -> int: ...
def ssl_session_print(bio: C.BIO, sess: C.SSL_SESSION) -> int: ...
def ssl_session_read_pem(bio: C.BIO) -> Optional[C.SSL_SESSION]: ...
def ssl_session_set_timeout(sess: C.SSL_SESSION, t: int) -> int: ...
def ssl_session_set_time(sess: C.SSL_SESSION, t: int) -> int: ...
def ssl_session_write_pem(sess: C.SSL_SESSION, bio: C.BIO) -> int: ...
def ssl_set_accept_state(ssl: C.SSL) -> None: ...
def ssl_set_bio(ssl: C.SSL, rbio: C.BIO, wbio: C.BIO) -> None: ...
def ssl_set_cipher_list(ssl: C.SSL, ciphers: str) -> int: ...
def ssl_set_client_CA_list_from_context(ssl: C.SSL, ctx: C.SSL_CTX) -> None: ...
def ssl_set_client_CA_list_from_file(ssl: C.SSL, ca_file: str) -> None: ...
def ssl_set_connect_state(ssl: C.SSL) -> None: ...
def ssl_set_session_id_context(ssl: C.SSL, sid_ctx: bytes) -> int: ...
def ssl_set_session(ssl: C.SSL, sess: C.SSL_SESSION) -> int: ...
def ssl_set_shutdown1(ssl: C.SSL, mode: int) -> None: ...
def ssl_set_tlsext_host_name(ssl: C.SSL, name: str) -> int: ...
def ssl_shutdown(ssl: C.SSL) -> int: ...
def sslv23_method() -> C.SSL_METHOD: ...
def ssl_write_nbio(ssl: C.SSL, data: bytes) -> int: ...
def ssl_write(ssl: C.SSL, data: bytes, timeout: float = -1.0) -> int: ...
def set_client_CA_list_from_file(ssl: C.SSL, ca_file: str) -> None: ...
def tlsv1_method() -> C.SSL_METHOD: ...

# --- X509 ---
def d2i_x509(bio: C.BIO) -> Optional[C.X509]: ...
def d2i_x509_req(bio: C.BIO) -> Optional[C.X509_REQ]: ...
def get_der_encoding_stack(stack: C.STACK_OF_X509) -> bytes: ...
def i2d_x509_bio(bio: C.BIO, x509: C.X509) -> int: ...
def i2d_x509_req_bio(bio: C.BIO, req: C.X509_REQ) -> int: ...
def i2d_x509(x: C.X509) -> bytes: ...
def make_stack_from_der_sequence(der: bytes) -> C.STACK_OF_X509: ...
def sk_x509_extension_free(stack: C.STACK_OF_X509_EXTENSION) -> None: ...
def sk_x509_extension_new_null() -> C.STACK_OF_X509_EXTENSION: ...
def sk_x509_extension_num(stack: C.STACK_OF_X509_EXTENSION) -> int: ...
def sk_x509_extension_pop(stack: C.STACK_OF_X509_EXTENSION) -> C.X509_EXTENSION: ...
def sk_x509_extension_push(
    stack: C.STACK_OF_X509_EXTENSION, ext: C.X509_EXTENSION
) -> int: ...
def sk_x509_extension_value(
    stack: C.STACK_OF_X509_EXTENSION, idx: int
) -> C.X509_EXTENSION: ...
def sk_x509_free(stack: C.STACK_OF_X509) -> None: ...
def sk_x509_new_null() -> C.STACK_OF_X509: ...
def sk_x509_num(stack: C.STACK_OF_X509) -> int: ...
def sk_x509_pop(stack: C.STACK_OF_X509) -> Optional[C.X509]: ...
def sk_x509_push(stack: C.STACK_OF_X509, cert: C.X509) -> int: ...
def sk_x509_value(stack: C.STACK_OF_X509, idx: int) -> C.X509: ...
def x509_add_ext(x: C.X509, ext: C.X509_EXTENSION, loc: int) -> int: ...
def x509_crl_free(crl: C.X509_CRL) -> None: ...
def x509_crl_new() -> C.X509_CRL: ...
def x509_crl_print(bio: C.BIO, crl: C.X509_CRL) -> int: ...
def x509_crl_read_pem(bio: C.BIO) -> Optional[C.X509_CRL]: ...
def x509_extension_free(ext: C.X509_EXTENSION) -> None: ...
def x509_extension_get_critical(ext: C.X509_EXTENSION) -> int: ...
def x509_extension_get_name(ext: C.X509_EXTENSION) -> bytes: ...
def x509_extension_set_critical(ext: C.X509_EXTENSION, crit: int) -> int: ...
def x509_ext_print(
    bio: C.BIO, ext: C.X509_EXTENSION, flags: int, indent: int
) -> int: ...
def x509_free(x: C.X509) -> None: ...
def x509_get_ext_count(x: C.X509) -> int: ...
def x509_get_ext(x: C.X509, loc: int) -> C.X509_EXTENSION: ...
def x509_get_issuer_name(x: C.X509) -> C.X509_NAME: ...
def x509_get_not_after(x509: C.X509) -> C.ASN1_Time: ...
def x509_get_not_before(x509: C.X509) -> C.ASN1_Time: ...
def x509_get_pubkey(x: C.X509) -> Optional[C.EVP_PKEY]: ...
def x509_get_serial_number(x: C.X509) -> C.ASN1_Integer: ...
def x509_get_subject_name(x: C.X509) -> C.X509_NAME: ...
def x509_get_version(x: C.X509) -> int: ...
def x509_check_ca(x: C.X509) -> int: ...
def x509_check_purpose(x: C.X509, id: int, ca: int) -> int: ...
def x509_init(err: Any) -> None: ...
def x509_name_add_entry_by_txt(
    name: C.X509_NAME,
    field: str,
    type: int,
    value: str,
    length: int,
    loc: int,
    set: int,
) -> int: ...
def x509_name_by_nid(name: C.X509_NAME, nid: int) -> Optional[bytes]: ...
def x509_name_entry_count(name: C.X509_NAME) -> int: ...
def x509_name_entry_create_by_txt(
    ne: C.X509_NAME_ENTRY, field: str, type: int, bytes: bytes, len: int
) -> C.X509_NAME_ENTRY: ...
def x509_name_entry_free(entry: C.X509_NAME_ENTRY) -> None: ...
def x509_name_entry_get_data(entry: C.X509_NAME_ENTRY) -> C.ASN1_String: ...
def x509_name_entry_get_object(entry: C.X509_NAME_ENTRY) -> C.ASN1_Object: ...
def x509_name_entry_set_data(
    entry: C.X509_NAME_ENTRY, type: int, data: bytes
) -> int: ...
def x509_name_entry_set_object(entry: C.X509_NAME_ENTRY, obj: C.ASN1_Object) -> int: ...
def x509_name_free(name: C.X509_NAME) -> None: ...
def x509_name_get_der(name: C.X509_NAME) -> bytes: ...
def x509_name_get_entry(name: C.X509_NAME, loc: int) -> C.X509_NAME_ENTRY: ...
def x509_name_get_index_by_nid(name: C.X509_NAME, nid: int, lastpos: int) -> int: ...
def x509_name_hash(name: C.X509_NAME) -> int: ...
def x509_name_new() -> C.X509_NAME: ...
def x509_name_oneline(name: C.X509_NAME) -> str: ...
def x509_name_print_ex(
    bio: C.BIO, name: C.X509_NAME, indent: int, flags: int
) -> int: ...
def x509_name_set_by_nid(name: C.X509_NAME, nid: int, value: str) -> int: ...
def x509_name_type_check(name: C.X509_NAME) -> int: ...
def x509_new() -> C.X509: ...
def x509_print(bio: C.BIO, x509: C.X509) -> int: ...
def x509_read_pem(bio: C.BIO) -> Optional[C.X509]: ...
def x509_req_add_extensions(
    req: C.X509_REQ, exts: C.STACK_OF_X509_EXTENSION
) -> int: ...
def x509_req_free(req: C.X509_REQ) -> None: ...
def x509_req_get_pubkey(req: C.X509_REQ) -> Optional[C.EVP_PKEY]: ...
def x509_req_get_subject_name(req: C.X509_REQ) -> C.X509_NAME: ...
def x509_req_get_version(req: C.X509_REQ) -> int: ...
def x509_req_new() -> C.X509_REQ: ...
def x509_req_print(bio: C.BIO, req: C.X509_REQ) -> int: ...
def x509_req_read_pem(bio: C.BIO) -> Optional[C.X509_REQ]: ...
def x509_req_set_pubkey(req: C.X509_REQ, pkey: C.EVP_PKEY) -> int: ...
def x509_req_set_subject_name(req: C.X509_REQ, name: C.X509_NAME) -> int: ...
def x509_req_set_version(req: C.X509_REQ, ver: int) -> int: ...
def x509_req_sign(req: C.X509_REQ, pkey: C.EVP_PKEY, md: C.EVP_MD) -> int: ...
def x509_req_verify(req: C.X509_REQ, pkey: C.EVP_PKEY) -> int: ...
def x509_req_write_pem(bio: C.BIO, req: C.X509_REQ) -> int: ...
def x509_set_issuer_name(x: C.X509, name: C.X509_NAME) -> int: ...
def x509_set_not_after(x509: C.X509, tm: C.ASN1_Time) -> int: ...
def x509_set_not_before(x509: C.X509, tm: C.ASN1_Time) -> int: ...
def x509_set_pubkey(x: C.X509, pkey: C.EVP_PKEY) -> int: ...
def x509_set_serial_number(x: C.X509, bn: C.ASN1_Integer) -> int: ...
def x509_set_subject_name(x: C.X509, name: C.X509_NAME) -> int: ...
def x509_set_version(x: C.X509, version: int) -> int: ...
def x509_sign(x: C.X509, pkey: C.EVP_PKEY, md: C.EVP_MD) -> int: ...
def x509_store_add_cert(store: C.X509_STORE, x: C.X509) -> int: ...
def x509_store_ctx_free(ctx: C.X509_STORE_CTX) -> None: ...
def x509_store_ctx_get_current_cert(ctx: C.X509_STORE_CTX) -> C.X509: ...
def x509_store_ctx_get_error(ctx: C.X509_STORE_CTX) -> int: ...
def x509_store_ctx_get_error_depth(ctx: C.X509_STORE_CTX) -> int: ...
def x509_store_ctx_get1_chain(ctx: C.X509_STORE_CTX) -> C.STACK_OF_X509: ...
def x509_store_free(store: C.X509_STORE) -> None: ...
def x509_store_load_locations(store: C.X509_STORE, cafile: str) -> int: ...
def x509_store_new() -> C.X509_STORE: ...
def x509_store_set_flags(store: C.X509_STORE, flags: int) -> int: ...
def x509_store_ctx_set_error(ctx: C.X509_STORE_CTX, err: int) -> None: ...
def x509_store_ctx_get_ex_data(ctx: C.X509_STORE_CTX, idx: int) -> C.SSL: ...
def x509_store_set_verify_cb(
    store: C.X509_STORE, cb: Callable[[int, C.X509_STORE_CTX], int]
) -> None: ...
def x509_type_check(x509: C.X509) -> int: ...
def x509_verify(x: C.X509, pkey: C.EVP_PKEY) -> int: ...
def x509v3_ext_conf(
    conf: Any, ctx: C.X509V3_CTX, name: str, value: str
) -> C.X509_EXTENSION: ...
def x509v3_set_nconf() -> C.X509V3_CTX: ...
def x509_write_pem(bio: C.BIO, x: C.X509) -> int: ...
def x509v3_ctx_free(ctx: C.X509V3_CTX) -> None: ...
def X509V3_CTX_set_nconf_pkey(ctx: C.X509V3_CTX, pkey: C.EVP_PKEY) -> None: ...

# --- S/MIME and PKCS7 ---
def smime_init(err: Any) -> None: ...
def pkcs7_init(err: Any) -> None: ...
def pkcs7_new() -> C.PKCS7: ...
def pkcs7_free(p7: C.PKCS7) -> None: ...
def pkcs7_sign0(
    signer_cert: C.X509,
    signer_key: C.EVP_PKEY,
    data: C.BIO,
    hash_obj: C.EVP_MD,
    flags: int,
) -> Optional[C.PKCS7]: ...
def pkcs7_sign1(
    signer_cert: C.X509,
    signer_key: C.EVP_PKEY,
    certs: C.STACK_OF_X509,
    data: C.BIO,
    hash_obj: C.EVP_MD,
    flags: int,
) -> Optional[C.PKCS7]: ...
def pkcs7_verify0(
    p7: C.PKCS7, certs: C.STACK_OF_X509, store: C.X509_STORE, flags: int
) -> bytes: ...
def pkcs7_verify1(
    p7: C.PKCS7, certs: C.STACK_OF_X509, store: C.X509_STORE, data: C.BIO, flags: int
) -> bytes: ...
def pkcs7_encrypt(
    certs: C.STACK_OF_X509, data: C.BIO, cipher: C.EVP_CIPHER, flags: int
) -> Optional[C.PKCS7]: ...
def pkcs7_decrypt(p7: C.PKCS7, pkey: C.EVP_PKEY, cert: C.X509, flags: int) -> bytes: ...
def pkcs7_read_bio(bio: C.BIO) -> Optional[C.PKCS7]: ...
def pkcs7_read_bio_der(bio: C.BIO) -> Optional[C.PKCS7]: ...
def pkcs7_write_bio(p7: C.PKCS7, bio: C.BIO) -> int: ...
def pkcs7_write_bio_der(p7: C.PKCS7, bio: C.BIO) -> int: ...
def smime_read_pkcs7(bio: C.BIO) -> Tuple[Optional[C.PKCS7], Optional[C.BIO]]: ...
def smime_write_pkcs7(bio: C.BIO, p7: C.PKCS7, flags: int) -> int: ...
def smime_write_pkcs7_multi(
    bio: C.BIO, p7: C.PKCS7, data: C.BIO, flags: int
) -> int: ...
def smime_crlf_copy(in_bio: C.BIO, out_bio: C.BIO) -> int: ...
def pkcs7_get0_signers(
    p7: C.PKCS7, certs: C.STACK_OF_X509, flags: int
) -> C.STACK_OF_X509: ...
def pkcs7_type_nid(p7: C.PKCS7) -> int: ...
def pkcs7_type_sn(p7: C.PKCS7) -> str: ...
def pkcs7_add_certificate(p7: C.PKCS7, cert: C.X509) -> None: ...

# --- Rand ---
def rand_init(err: Any) -> None: ...
def rand_add(blob: bytes, entropy: float) -> None: ...
def rand_seed(seed: bytes) -> None: ...
def rand_status() -> int: ...
def rand_file_name() -> bytes: ...
def rand_load_file(filename: str, max_bytes: int) -> int: ...
def rand_save_file(filename: str) -> int: ...
def rand_bytes(n: int) -> bytes: ...
def rand_pseudo_bytes(n: int) -> Tuple[bytes, int]: ...



"""M2Crypto wrapper for OpenSSL DH API.

Copyright (c) 1999-2003 Ng Pheng Siong. All rights reserved."""

from M2Crypto import BIO, m2, types as C
from M2Crypto.util import genparam_callback
from typing import Callable, Optional, Union  # noqa


class DHError(Exception):
    pass


m2.dh_init(DHError)


class DH(object):
    """Object interface to the Diffie-Hellman key exchange protocol."""

    def __init__(self, dh: C.DH, _pyfree: int = 0) -> None:
        assert m2.dh_type_check(dh)
        self.dh = dh
        self._pyfree = _pyfree

    @staticmethod
    def m2_dh_free(dh: C.DH) -> None:
        m2.dh_free(dh)

    def __del__(self) -> None:
        if getattr(self, "_pyfree", 0):
            self.m2_dh_free(self.dh)

    def __len__(self) -> int:
        assert m2.dh_type_check(self.dh), "'dh' type error"
        return int(m2.dh_size(self.dh))

    def __getattr__(self, name: str) -> bytes:
        if name in ("p", "g", "pub", "priv"):
            method = getattr(m2, "dh_get_%s" % (name,))
            assert m2.dh_type_check(self.dh), "'dh' type error"
            return method(self.dh)
        else:
            raise AttributeError

    def __setattr__(self, name: str, value: bytes) -> None:
        if name in ("p", "g"):
            raise DHError("set (p, g) via set_params()")
        elif name in ("pub", "priv"):
            raise DHError("generate (pub, priv) via gen_key()")
        else:
            self.__dict__[name] = value

    def _ptr(self):
        return self.dh

    def check_params(self) -> int:
        assert m2.dh_type_check(self.dh), "'dh' type error"
        return m2.dh_check(self.dh)

    def gen_key(self) -> None:
        assert m2.dh_type_check(self.dh), "'dh' type error"
        m2.dh_generate_key(self.dh)

    def compute_key(self, pubkey: bytes) -> bytes:
        assert m2.dh_type_check(self.dh), "'dh' type error"
        return m2.dh_compute_key(self.dh, pubkey)

    def print_params(self, bio: BIO.BIO) -> int:
        assert m2.dh_type_check(self.dh), "'dh' type error"
        return m2.dhparams_print(bio._ptr(), self.dh)


def gen_params(
    plen: int,
    g: int,
    callback: Callable = genparam_callback,
) -> DH:
    dh_parms = m2.dh_generate_parameters(plen, g, callback)
    dh_obj = DH(dh_parms, 1)
    return dh_obj


def load_params(file: Union[str, bytes]) -> DH:
    with BIO.openfile(file) as bio:
        return load_params_bio(bio)


def load_params_bio(bio: BIO.BIO) -> DH:
    dh_ptr = m2.dh_read_parameters(bio._ptr())
    if dh_ptr is None:
        raise DHError("Failed to load DH parameters.")
    return DH(dh_ptr, 1)


def set_params(p: bytes, g: bytes) -> DH:
    dh = m2.dh_new()
    m2.dh_set_pg(dh, p, g)
    return DH(dh, 1)


# def free_params(cptr):
#    m2.dh_free(cptr)


DH_GENERATOR_2 = m2.DH_GENERATOR_2
DH_GENERATOR_5 = m2.DH_GENERATOR_5



"""M2Crypto wrapper for OpenSSL RC4 API.

Copyright (c) 1999-2003 Ng Pheng Siong. All rights reserved."""

from typing import Optional
from M2Crypto import m2, types as C


class RC4(object):
    """Object interface to the stream cipher RC4."""

    def __init__(self, key: Optional[bytes] = None) -> None:
        self.cipher = m2.rc4_new()
        if key:
            m2.rc4_set_key(self.cipher, key)

    @staticmethod
    def rc4_free(cipher: C.RC4_KEY) -> None:
        m2.rc4_free(cipher)

    def __del__(self) -> None:
        if getattr(self, "cipher", None):
            self.rc4_free(self.cipher)

    def set_key(self, key: bytes) -> None:
        m2.rc4_set_key(self.cipher, key)

    def update(self, data: bytes) -> bytes:
        return m2.rc4_update(self.cipher, data)

    def final(self) -> str:
        return ""



"""
M2Crypto is the most complete Python wrapper for OpenSSL
featuring RSA, DSA, DH, EC, HMACs, message digests, symmetric
ciphers (including AES); SSL functionality to implement clients
and servers; HTTPS extensions to Python's httplib, urllib, and
xmlrpclib; unforgeable HMAC'ing AuthCookies for web session
management; FTP/TLS client and server; and ZSmime: An S/MIME
messenger for Zope.  M2Crypto can also be used to provide SSL for
Twisted.

Copyright (c) 1999-2004 Ng Pheng Siong. All rights reserved.

Portions created by Open Source Applications Foundation (OSAF) are
Copyright (C) 2004-2007 OSAF. All Rights Reserved.

Copyright 2008-2011 Heikki Toivonen. All rights reserved.
"""

__version__: str = "0.46.2"
version: str = __version__

try:
    from packaging.version import Version
except ImportError:
    try:
        from distutils.version import StrictVersion as Version  # type: ignore [no-redef, import-not-found]
    except ImportError:
        Version = None  # type: ignore[misc, assignment]

if Version is not None:
    version_info: tuple = (0, 0, 0)
    __ver: Version = Version(__version__)
    if hasattr(__ver, "_version"):
        version_info = tuple(__ver._version[1])
    elif hasattr(__ver, "version"):
        version_info = __ver.version

from M2Crypto import m2

encrypt: int = 1
decrypt: int = 0
